# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

"""densenet sub-module within the ketos.neural_networks module

    This module provides classes to implement Dense Networks (DenseNets).

    Contents: ConvBlock class
              DenseBlock class
              TransitionBlock class
              DenseNetArch class
              DenseNetInterface class
        
"""

import tensorflow as tf
from ketos.neural_networks.dev_utils.nn_interface import RecipeCompat, NNInterface, NNArch
import json


default_densenet_recipe =  {'dense_blocks':[ 6, 12, 24, 16],
                    'growth_rate':32,
                    'compression_factor':0.5,
                    'n_classes':2,
                    'dropout_rate':0.2,
                    'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                    'loss_function': RecipeCompat('CategoricalCrossentropy', tf.keras.losses.CategoricalCrossentropy),  
                    'metrics': [RecipeCompat('BinaryAccuracy',tf.keras.metrics.BinaryAccuracy),
                                RecipeCompat('Precision',tf.keras.metrics.Precision, class_id=1),
                                RecipeCompat('Recall',tf.keras.metrics.Recall, class_id=1)],
                    }





class ConvBlock(tf.keras.Model):
    """ Convolutional Blocks used in the Dense Blocks.

        Args:
            growth_rate:int
                The growth rate for the number of filters (i.e.: channels) between convolutional layers
    
    """
    def __init__(self, growth_rate):
        super(ConvBlock, self).__init__()

        self.growth_rate = growth_rate

        self.batch_norm1 = tf.keras.layers.BatchNormalization(epsilon=1.001e-5)
        self.relu1 = tf.keras.layers.Activation('relu')
        self.conv1 = tf.keras.layers.Conv2D(4 * self.growth_rate, kernel_size=1, strides=1, use_bias=False, padding="same")

        self.batch_norm2 = tf.keras.layers.BatchNormalization(epsilon=1.001e-5)
        self.relu2 = tf.keras.layers.Activation('relu')
        self.conv2 = tf.keras.layers.Conv2D(self.growth_rate, kernel_size=3, strides=1, use_bias=False, padding="same")

    def call(self, inputs, training=False):
        """Calls the model on new inputs.

        In this case call just reapplies all ops in the graph to the new inputs (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: Tensor or list of tensors
                A tensor or list of tensors
            
            training: Bool
                Boolean or boolean scalar tensor, indicating whether to run the Network in training mode or inference mode.

        Returns:
                A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        outputs = self.batch_norm1(inputs, training=training)
        outputs = self.relu1(outputs)
        outputs = self.conv1(outputs)

        outputs = self.batch_norm2(outputs, training=training)
        outputs = self.relu2(outputs)
        outputs = self.conv2(outputs)

        return outputs




class DenseBlock(tf.keras.Model):
    """ Dense block for DenseNet architectures

        Args:
            growth_rate: int
                The growth rate between blocks
            n_blocks:
                The number of convolutional blocks within the dense block
    
    """
    def __init__(self, growth_rate, n_blocks):
        super(DenseBlock,self).__init__()

        self.n_blocks = n_blocks
        self.blocks = tf.keras.Sequential()
        for i_block in range(self.n_blocks):
            self.blocks.add(ConvBlock(growth_rate=growth_rate))
        

    def call(self, inputs, training=False):
        """Calls the model on new inputs.

        In this case call just reapplies all ops in the graph to the new inputs (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: Tensor or list of tensors
                A tensor or list of tensors
            
            training: Bool
                Boolean or boolean scalar tensor, indicating whether to run the Network in training mode or inference mode.

        Returns:
                A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        outputs = self.blocks(inputs, training=training)
        outputs = tf.keras.layers.concatenate([inputs, outputs])

        return outputs


class TransitionBlock(tf.keras.Model):
    """ Transition Blocks for the DenseNet architecture

        Args:
            n_filters:int
                Number of filters (i,e,: channels)
            compression_factor: float
                The compression factor used within the transition block
                (i.e.: the reduction of filters/channels from the previous dense block to the next)
            dropout_rate:float
                Dropout rate for the convolutional layer (between 0 and 1, use 0 for no dropout)

    """
    def __init__(self, n_channels, compression_factor, dropout_rate=0.2):
        super(TransitionBlock, self).__init__()
        
        self.n_channels = n_channels
        self.compression_factor = compression_factor
        self.dropout_rate = dropout_rate

        self.batch_norm = tf.keras.layers.BatchNormalization(epsilon=1.001e-5)
        self.conv = tf.keras.layers.Conv2D(int(self.n_channels * self.compression_factor), kernel_size=1, strides=1, padding="same")
        self.dropout = tf.keras.layers.Dropout(self.dropout_rate)
        self.relu = tf.keras.layers.Activation('relu')
        self.avg_pool = tf.keras.layers.AveragePooling2D((2,2), strides=2)

    def call(self, inputs, training=False):
        """Calls the model on new inputs.

        In this case call just reapplies all ops in the graph to the new inputs (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: Tensor or list of tensors
                A tensor or list of tensors
            
            training: Bool
                Boolean or boolean scalar tensor, indicating whether to run the Network in training mode or inference mode.

        Returns:
                A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        outputs = self.batch_norm(inputs, training=training)
        outputs = self.relu(outputs)
        outputs = self.conv(outputs)
        outputs = self.dropout(outputs, training=training)
        outputs = self.avg_pool(outputs)

        return outputs


class DenseNetArch(NNArch):
    """Implements a DenseNet architecture, building on top of Dense and tansition blocks

        Args:
            block_sets: list of ints
                A list specifying the block sets and how many blocks each set contains.
                Example: [6, 12, 24, 16]  will create a DenseNet with 4 block sets containing 6, 12, 24 and 16
                dense blocks, with a total of 58 blocks.
            growth_rate:int
                The factor by which the number of filters (i.e.: channels) within each dense block grows.
            compression_factor: float
                The factor by which transition blocks reduce the number of filters (i.e.: channels) between dense blocks (between 0 and 1).
            dropout_rate: float
                The droput rate (between 0 and 1) used in each transition block. Use 0 for no dropout.
            n_classes:int
                The number of classes. The output layer uses a Softmax activation and
                will contain this number of nodes, resulting in model outputs with this
                many values summing to 1.0.
            pre_trained_base: instance of DenseNetArch
                A pre-trained densenet model from which the residual blocks will be taken. 
                Use by the the clone_with_new_top method when creating a clone for transfer learning
        """

        

    def __init__(self, dense_blocks, growth_rate, compression_factor, n_classes, dropout_rate, pre_trained_base=None):
        super(DenseNetArch, self).__init__()

        self.dense_blocks = dense_blocks
        self.growth_rate = growth_rate
        self.compression_factor = compression_factor
        self.n_classes = n_classes
        self.dropout_rate = dropout_rate

        if pre_trained_base:
            self.initial_conv = pre_trained_base[0]
            self.initial_batch_norm = pre_trained_base[1]
            self.initial_relu = pre_trained_base[2]
            self.initial_pool = pre_trained_base[3]

            self.dense_blocks_seq = pre_trained_base[4]

        else:
            self.initial_conv = tf.keras.layers.Conv2D(2 * self.growth_rate, kernel_size=7, strides=2, padding="same")
            self.initial_batch_norm = tf.keras.layers.BatchNormalization(epsilon=1.001e-5)
            self.initial_relu = tf.keras.layers.Activation('relu')
            self.initial_pool = tf.keras.layers.MaxPool2D((2,2), strides=2)

            self.n_channels = 2 * self.growth_rate
            self.dense_blocks_seq = tf.keras.Sequential()
            for n_layers in self.dense_blocks:
                self.dense_blocks_seq.add(DenseBlock(growth_rate=self.growth_rate, n_blocks=n_layers))
                self.n_channels += n_layers * self.growth_rate
                self.dense_blocks_seq.add(TransitionBlock(n_channels=self.n_channels, compression_factor=self.compression_factor, dropout_rate=self.dropout_rate))

        self.global_avg_pool = tf.keras.layers.GlobalAveragePooling2D()
        self.flatten = tf.keras.layers.Flatten()
        self.dense = tf.keras.layers.Dense(self.n_classes)
        self.softmax = tf.keras.layers.Softmax()

    
    def call(self, inputs, training=False):
        """Calls the model on new inputs.

        In this case call just reapplies all ops in the graph to the new inputs (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: Tensor or list of tensors
                A tensor or list of tensors
            
            training: Bool
                Boolean or boolean scalar tensor, indicating whether to run the Network in training mode or inference mode.

        Returns:
                A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        outputs = self.call_frontend(inputs)
        outputs = self.initial_conv(outputs)
        outputs = self.initial_batch_norm(outputs, training=training)
        outputs = self.initial_relu(outputs)
        outputs = self.initial_pool(outputs)

        outputs = self.dense_blocks_seq(outputs)

        outputs = self.global_avg_pool(outputs)
        outputs = self.flatten(outputs)
        outputs = self.dense(outputs)
        outputs = self.softmax(outputs)

        return outputs


    def freeze_init_layer(self):
        """Freeze the initial convolutional layer"""
        self.layers[0].trainable = False

    def unfreeze_init_layer(self):
        """Unfreeze the initial convolutional layer"""
        self.layers[0].trainable = True
    
    def freeze_block(self, block_ids):
        """ Freeze specific dense blocks

            Args:
                blocks_ids: list of ints
                    The block numbers to be freezed (starting from zero)
        """

        for block_id in block_ids:
            self.layers[4].layers[block_id].trainable = False

    def unfreeze_block(self, block_ids):
        """ Unfreeze specific dense blocks

             Args:
                blocks_ids: list of ints
                    The block numbers to be freezed (starting from zero)
        """
        for block_id in block_ids:
            self.layers[4].layers[block_id].trainable = True
    
    def freeze_top(self):
        """Freeze the classification block"""
        for layer in self.layers[5:]:
            layer.trainable = False
    
    def unfreeze_top(self):
        """Unfreeze the classification block"""
        for layer in self.layers[5:]:
            layer.trainable = True


    def get_feature_extraction_base(self):
        """ Retrive the feature extraction base (initial convolutional layer + dense blocks)
        
            Returns:
                list containing the feature extraction layers
        """
        return [ self.initial_conv,
                self.initial_batch_norm,
                self.initial_relu,
                self.initial_pool,
                self.dense_blocks_seq]

    def clone_with_new_top(self, n_classes=None, freeze_base=True):
        """ Clone this instance but replace the original classification top with a new (untrained) one
        
            Args:
                n_classes:int
                    The number of classes the new classification top should output.
                    If None(default), the original number of classes will be used.
                freeze_base:bool
                    If True, the weights of the feature extraction base will be froze (untrainable) in the new model.
                
            Returns:
                cloned_model: instance of DenseNetArch
                    The new model with the old feature extraction base and new classification top.
         """
        if freeze_base == True:
            self.trainable = False

        if n_classes is None:
            n_classes = self.n_classes

        pre_trained_base = self.get_feature_extraction_base()
        cloned_model = type(self)(n_classes=n_classes, pre_trained_base=pre_trained_base)

        return cloned_model


class DenseNetInterface(NNInterface):


    @classmethod
    def _build_from_recipe(cls, recipe, recipe_compat=True):
        """ Build a DenseNet model from a recipe.

            Args:
                recipe: dict
                    A recipe dictionary. optimizer, loss function
                    and metrics must be instances of ketos.neural_networks.RecipeCompat.
                    
                    Example recipe:
                    
                    >>> {'dense_blocks': [6, 12, 24, 16], # doctest: +SKIP
                    ...    'growth_rate':32,
                    ...    'compression_factor':0.5,
                    ...    'n_classes':2,
                    ...    'dropout_rate':0.2,        
                    ...    'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                    ...    'loss_function': RecipeCompat('BinaryCrossentropy', tf.keras.losses.BinaryCrossentropy),  
                    ...    'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)],
                    }
                     

            Returns:
                An instance of ResNetInterface.

        """

        dense_blocks = recipe['dense_blocks']
        growth_rate = recipe['growth_rate']
        compression_factor = recipe['compression_factor']
        n_classes = recipe['n_classes']
        dropout_rate = recipe['dropout_rate']
        
        
        if recipe_compat == True:
            optimizer = recipe['optimizer']
            loss_function = recipe['loss_function']
            metrics = recipe['metrics']
            
        else:
            optimizer = cls._optimizer_from_recipe(recipe['optimizer'])
            loss_function = cls._loss_function_from_recipe(recipe['loss_function'])
            metrics = cls._metrics_from_recipe(recipe['metrics'])
            

        instance = cls(dense_blocks=dense_blocks, growth_rate=growth_rate, compression_factor=compression_factor, n_classes=n_classes, dropout_rate=dropout_rate, optimizer=optimizer, loss_function=loss_function, metrics=metrics)

        return instance

    @classmethod
    def _read_recipe_file(cls, json_file, return_recipe_compat=True):
        """ Read a DenseNet recipe saved in a .json file.

            Args:
                json_file:string
                    Full path (including filename and extension) to the .json file containing the recipe.
                return_recipe_compat:bool
                    If True, returns a dictionary where the optimizer, loss_function, metrics and 
                    secondary_metrics (if available) values are instances of the ketos.neural_networks.nn_interface.RecipeCompat.
                    The returned dictionary will be equivalent to:
                            
                            >>> {'dense_blocks': [6, 12, 24, 16], # doctest: +SKIP
                            ...  'growth_rate':32,
                            ...  'compression_factor':0.5,
                            ...  'n_classes':2,
                            ...  'dropout_rate':0.2,        
                            ...  'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                            ...  'loss_function': RecipeCompat('BinaryCrossentropy', tf.keras.losses.BinaryCrossentropy),  
                            ...  'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)],
                            }

                    If False, the optimizer, loss_function, metrics and secondary_metrics (if available) values will contain a
                    dictionary representation of such fields instead of the RecipeCompat objects:
                            >>> {'dense_blocks': [6, 12, 24, 16], # doctest: +SKIP
                            ...  'growth_rate':32,
                            ...  'compression_factor':0.5,
                            ...  'n_classes':2,
                            ...  'dropout_rate':0.2,        
                            ...  'optimizer': {'name':'Adam', 'parameters': {'learning_rate':0.005}},
                            ...  'loss_function': {'name':'BinaryCrossentropy', 'parameters':{}},  
                            ...  'metrics': [{'name':'CategoricalAccuracy', 'parameters':{}}]}

                Returns:
                    recipe, according to 'return_recipe_compat.

        """

        with open(json_file, 'r') as json_recipe:
            recipe_dict = json.load(json_recipe)

        optimizer = cls._optimizer_from_recipe(recipe_dict['optimizer'])
        loss_function = cls._loss_function_from_recipe(recipe_dict['loss_function'])
        metrics = cls._metrics_from_recipe(recipe_dict['metrics'])
        
        if return_recipe_compat == True:
            recipe_dict['optimizer'] = optimizer
            recipe_dict['loss_function'] = loss_function
            recipe_dict['metrics'] = metrics
            
        else:
            recipe_dict['optimizer'] = cls._optimizer_to_recipe(optimizer)
            recipe_dict['loss_function'] = cls._loss_function_to_recipe(loss_function)
            recipe_dict['metrics'] = cls._metrics_to_recipe(metrics)
        
        # recipe_dict['block_sets'] = recipe_dict['block_sets']
        # recipe_dict['n_classes'] = recipe_dict['n_classes']
        # recipe_dict['initial_filters'] = recipe_dict['initial_filters']

        return recipe_dict

    def __init__(self, dense_blocks=default_densenet_recipe['dense_blocks'], growth_rate=default_densenet_recipe['growth_rate'], compression_factor=default_densenet_recipe['compression_factor'], n_classes=default_densenet_recipe['n_classes'], dropout_rate=default_densenet_recipe['dropout_rate'], optimizer=default_densenet_recipe['optimizer'], loss_function=default_densenet_recipe['loss_function'], metrics=default_densenet_recipe['metrics']):
        super(DenseNetInterface, self).__init__(optimizer=optimizer, loss_function=loss_function, metrics=metrics)

        self.dense_blocks = dense_blocks
        self.growth_rate = growth_rate
        self.compression_factor = compression_factor
        self.n_classes = n_classes
        self.dropout_rate = dropout_rate

        self.model = DenseNetArch(dense_blocks = self.dense_blocks, growth_rate=self.growth_rate, compression_factor=self.compression_factor, n_classes=self.n_classes, dropout_rate=self.dropout_rate)

    def _extract_recipe_dict(self):
        """ Create a recipe dictionary from a DenseNetInterface instance.

            The resulting recipe contains all the fields necessary to build the same network architecture used by the instance calling this method.
            
            Returns:
                recipe:dict
                    A dictionary containing the recipe fields necessary to build the same network architecture.
                    The output is equivalent to:
                       >>> {'dense_blocks': [6, 12, 24, 16], # doctest: +SKIP
                       ...  'growth_rate':32,
                       ...  'compression_factor':0.5,
                       ...  'n_classes':2,
                       ...  'dropout_rate':0.2,        
                       ...  'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                       ...  'loss_function': RecipeCompat('BinaryCrossentropy', tf.keras.losses.BinaryCrossentropy),  
                       ...  'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)],
                        }
      
        """

        recipe = {}
        recipe['interface'] = type(self).__name__
        recipe['dense_blocks'] = self.dense_blocks
        recipe['growth_rate'] = self.growth_rate
        recipe['compression_factor'] = self.compression_factor
        recipe['n_classes'] = self.n_classes
        recipe['dropout_rate'] = self.dropout_rate
        recipe['optimizer'] = self._optimizer_to_recipe(self.optimizer)
        recipe['loss_function'] = self._loss_function_to_recipe(self.loss_function)
        recipe['metrics'] = self._metrics_to_recipe(self.metrics)
        
        return recipe





# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" cnn sub-module within the ketos.neural_networks module

    This module provides classes to implement Convolutional Neural Networks (CNNs).

    Contents:
        CNN class:
        CNNInterface class:
"""
import tensorflow as tf
import numpy as np
from ketos.neural_networks.dev_utils.nn_interface import NNInterface, RecipeCompat, NNArch
import json



vgg_like_recipe = {'convolutional_layers':  [{'n_filters':64, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool': None, 'batch_normalization':True},
                                    {'n_filters':64, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool': {'pool_size':(2,2) , 'strides':(2,2)}, 'batch_normalization':True},
                                    {'n_filters':128, "filter_shape":(3,3), 'strides':1, 'padding':'valid','activation':'relu', 'max_pool':None, 'batch_normalization':True, },
                                    {'n_filters':128, "filter_shape":(3,3), 'strides':1, 'padding':'valid','activation':'relu', 'max_pool':{'pool_size':(2,2) , 'strides':(2,2)}, 'batch_normalization':True},
                                    {'n_filters':256, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True, },
                                    {'n_filters':256, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True, },
                                    {'n_filters':256, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True,},
                                    {'n_filters':256, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':(2,2) , 'strides':(2,2)}, 'batch_normalization':True},
                                    {'n_filters':512, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True, },
                                    {'n_filters':512, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True, },
                                    {'n_filters':512, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True, },
                                    {'n_filters':512, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':(2,2) , 'strides':(2,2)}, 'batch_normalization':True,},
                                    {'n_filters':512, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True,},
                                    {'n_filters':512, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True,},
                                    {'n_filters':512, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True,},
                                    {'n_filters':512, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':(2,2) , 'strides':(2,2)}, 'batch_normalization':True,}],
                 
                  'dense_layers':[{'n_hidden':4096, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                                    {'n_hidden':4096, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                                    ],
                  'n_classes': 2 ,
                  'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                  'loss_function': RecipeCompat('BinaryCrossentropy', tf.keras.losses.BinaryCrossentropy),  
                  'metrics': [RecipeCompat('BinaryAccuracy',tf.keras.metrics.BinaryAccuracy)]
                  

}


                                    


alexnet_like_recipe = {'convolutional_layers':  [{'n_filters':96, "filter_shape":(11,11), 'strides':4, 'padding':'valid',  'activation':'relu', 'max_pool': {'pool_size':(3,3) , 'strides':(2,2)}, 'batch_normalization':True, },
                                    {'n_filters':256, "filter_shape":(5,5), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool': {'pool_size':(3,3) , 'strides':(2,2)}, 'batch_normalization':True, },
                                    {'n_filters':384, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True,},
                                    {'n_filters':384, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True,},
                                    {'n_filters':256, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':(3,3) , 'strides':(2,2)}, 'batch_normalization':True,},],
                  
                  'dense_layers':[{'n_hidden':4096, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                                    {'n_hidden':4096, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                                    ],

                  'n_classes': 2 ,
                  'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                  'loss_function': RecipeCompat('BinaryCrossentropy', tf.keras.losses.BinaryCrossentropy),  
                  'metrics': [RecipeCompat('BinaryAccuracy',tf.keras.metrics.BinaryAccuracy)]                 


                    }

default_cnn_recipe = {'convolutional_layers':  [{'n_filters':32, "filter_shape":(8,8), 'strides':4, 'padding':'valid',  'activation':'relu', 'max_pool': {'pool_size':(3,3) , 'strides':(2,2)}, 'batch_normalization':True, },
                                    {'n_filters':64, "filter_shape":(3,3), 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool': {'pool_size':(3,3) , 'strides':(2,2)}, 'batch_normalization':True, },],
                  
                  'dense_layers':[{'n_hidden':512, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                                    {'n_hidden':128, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},],

                  'n_classes': 2 ,
                  'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                  'loss_function': RecipeCompat('BinaryCrossentropy', tf.keras.losses.BinaryCrossentropy),  
                  'metrics': [RecipeCompat('BinaryAccuracy',tf.keras.metrics.BinaryAccuracy)]
                  
                    }


default_cnn_1d_recipe = {'convolutional_layers':  [{'n_filters':8, "filter_shape":128, 'strides':2, 'padding':'causal', 'activation':'relu', 'max_pool': None, 'batch_normalization':True},
                                    {'n_filters':16, "filter_shape":64, 'strides':2, 'padding':'causal', 'activation':'relu', 'max_pool': {'pool_size': 8 , 'strides':8}, 'batch_normalization':True},
                                    {'n_filters':32, "filter_shape":32, 'strides':2, 'padding':'causal', 'activation':'relu', 'max_pool': {'pool_size': 8 , 'strides':8}, 'batch_normalization':True},
                                    {'n_filters':64, "filter_shape":16, 'strides':2, 'padding':'causal','activation':'relu', 'max_pool':None, 'batch_normalization':True, },
                                    {'n_filters':128, "filter_shape":8, 'strides':2, 'padding':'causal','activation':'relu', 'max_pool':None, 'batch_normalization':True},
                                    {'n_filters':256, "filter_shape":4, 'strides':2, 'padding':'causal', 'activation':'relu', 'max_pool':{'pool_size': 4 , 'strides': 4}, 'batch_normalization':True, },
                                    ],

                  'dense_layers':[{'n_hidden':512, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                                    {'n_hidden':128, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                                    ],

                  'n_classes': 2 ,
                  'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, lr=0.01, beta_1=0.9, beta_2=0.999, decay=0.01),
                  'loss_function': RecipeCompat('CategoricalCrossentropy', tf.keras.losses.CategoricalCrossentropy, from_logits=True),  
                  'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy),   
                              RecipeCompat('Precision',tf.keras.metrics.Precision, class_id=1),
                              RecipeCompat('Recall',tf.keras.metrics.Recall, class_id=1)],  
                    }


class CNNArch(NNArch):
    """ Implement a Convolutional Neural Network

        Note: in addition to the dense layers specified in the 'dense_layers' argument, an extra dense
              layer will always be added to the end. The output of this layer is determined by the 'n_classes'
              parameter. 

        Args:
            convolutional_layers: list
                A list of dictionaries containing the detailed specification for the convolutional layers.
                Each layer is specified as a dictionary with the following format:

                >>> {'n_filters':96, "filter_shape":(11,11), 'strides':4, 'padding':'valid', activation':'relu', 'max_pool': {'pool_size':(3,3) , 'strides':(2,2)}, 'batch_normalization':True} # doctest: +SKIP

            dense_layers: list
                A list of dictionaries containing the detailed specification for the fully connected layers.
                Each layer is specified as a dictionary with the following format:
                
                >>> {'n_hidden':4096, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5} # doctest: +SKIP
                
                This list should not include the output layr, which will be automatically added based on the 'n_classes' parameter.

             pre_trained_base: instance of CNNArch
                A pre-trained CNN model from which the residual blocks will be taken. 
                Use by the the clone_with_new_top method when creating a clone for transfer learning

            n_classes:int
                The number of classes the network will be used to classify.
                The output will be this number of values representing the scores for each class. 
                Scores sum to 1.0.
    """


    def __init__(self, convolutional_layers, dense_layers, n_classes, pre_trained_base=None, **kwargs):
        super(CNNArch, self).__init__(**kwargs)

        if pre_trained_base:
            self.convolutional_block = pre_trained_base[0]
        else:

            self.convolutional_block = tf.keras.models.Sequential(name="convolutional_block")
            for conv_layer in convolutional_layers:
                self.convolutional_block.add(tf.keras.layers.Conv2D(filters=conv_layer['n_filters'], kernel_size=conv_layer['filter_shape'], strides=conv_layer['strides'], activation=conv_layer['activation'], padding=conv_layer['padding']))
                if conv_layer['max_pool'] is not None:
                    self.convolutional_block.add(tf.keras.layers.MaxPooling2D(pool_size=conv_layer['max_pool']['pool_size'], strides=conv_layer['max_pool']['strides'] ))
                if conv_layer['batch_normalization'] == True:
                    self.convolutional_block.add(tf.keras.layers.BatchNormalization())
        
        self.flatten = tf.keras.layers.Flatten()

        self.dense_block = tf.keras.models.Sequential(name="dense_block")
        for fc_layer in dense_layers:
            self.dense_block.add(tf.keras.layers.Dense(units=fc_layer['n_hidden'], activation=fc_layer['activation']))
            if fc_layer['batch_normalization'] == True:
                self.dense_block.add(tf.keras.layers.BatchNormalization())
            if fc_layer['dropout'] > 0.0:
                self.dense_block.add(tf.keras.layers.Dropout(fc_layer['dropout']))

        
        self.dense_block.add(tf.keras.layers.Dense(n_classes))
        self.dense_block.add(tf.keras.layers.Softmax())

    
    def freeze_conv_block(self):
        """Freeze the convolutional block"""
        self.layers[0].trainable = False

    def unfreeze_conv_block(self):
        """Unfreeze the convolutional block"""
        self.layers[0].trainable = True
    
    def freeze_top(self):
        """Freeze the classification (dense) block"""
        for layer in self.layers[1:]:
            layer.trainable = False
    
    def unfreeze_top(self):
        """Unfreeze the classification (dense) block"""
        for layer in self.layers[1:]:
            layer.trainable = True

    def get_feature_extraction_base(self):
        """ Retrive the feature extraction base (initial convolutional layer + residual blocks)
        
            Returns:
                list containing the feature extraction layers
        """
        return [self.convolutional_block]

    def clone_with_new_top(self, n_classes=None, freeze_base=True):
        """ Clone this instance but replace the original classification top with a new (untrained) one
        
            Args:
                n_classes:int
                    The number of classes the new classification top should output.
                    If None(default), the original number of classes will be used.
                freeze_base:bool
                    If True, the weights of the feature extraction base will be froze (untrainable) in the new model.
                
            Returns:
                cloned_model: instance of CNNArch
                    The new model with the old feature extraction base and new classification top.
         """
        if freeze_base == True:
            self.trainable = False

        if n_classes is None:
            n_classes = self.n_classes

        pre_trained_base = self.get_feature_extraction_base()
        cloned_model = type(self)(n_classes=n_classes, pre_trained_base=pre_trained_base)

        return cloned_model

    def call(self, inputs, training=None):
        """Calls the model on new inputs.

        In this case call just reapplies all ops in the graph to the new inputs (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: Tensor or list of tensors
                A tensor or list of tensors
            training: Bool
                Boolean or boolean scalar tensor, indicating whether to run the Network in training mode or inference mode.

        Returns: 
            A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        output = self.call_frontend(inputs)
        output = self.convolutional_block(output, training=training)
        output = self.flatten(output)
        output = self.dense_block(output, training=training)
        
        return output




class CNN1DArch(NNArch):
    """ Implement an 1D (temporal) Convolutional Neural Network.

        Note: in addition to the dense layers specified in the 'dense_layers' argument, an extra dense \
              layer will always be added to the end. The output of this layer is determined by the 'n_classes' \
              parameter. 

        Args:
            convolutional_layers: list
                A list of dictionaries containing the detailed specification for the convolutional layers.
                Each layer is specified as a dictionary with the following format:

                >>> {'n_filters':96, "filter_shape":(11,11), 'strides':4, 'padding':'valid', activation':'relu', 'max_pool': {'pool_size':(3,3) , 'strides':(2,2)}, 'batch_normalization':True} # doctest: +SKIP

            dense_layers: list
                A list of dictionaries containing the detailed specification for the fully connected layers.
                Each layer is specified as a dictionary with the following format:

                >>> {'n_hidden':4096, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5} # doctest: +SKIP

                This list should not include the output layr, which will be automatically added based on the 'n_classes' parameter.

            n_classes:int
                The number of classes the network will be used to classify.
                The output will be this number of values representing the scores for each class. 
                Scores sum to 1.0.

            pre_trained_base: instance of CNN1DArch
                A pre-trained CNN 1D model from which the residual blocks will be taken. 
                Use by the the clone_with_new_top method when creating a clone for transfer learning
    """

    def __init__(self, dense_layers, n_classes, pre_trained_base=None, convolutional_layers=None, **kwargs): 

        self.convolutional_layers = convolutional_layers
        self.dense_layers = dense_layers
        self.n_classes = n_classes
        super(CNN1DArch, self).__init__(**kwargs)

        if pre_trained_base:
            self.convolutional_block = pre_trained_base[0]
        else:
            self.convolutional_block = tf.keras.models.Sequential(name="convolutional_block")
            for conv_layer in self.convolutional_layers:
                self.convolutional_block.add(tf.keras.layers.Conv1D(filters=conv_layer['n_filters'], kernel_size=conv_layer['filter_shape'], strides=conv_layer['strides'], activation=conv_layer['activation'], padding=conv_layer['padding']))
                if conv_layer['max_pool'] is not None:
                    self.convolutional_block.add(tf.keras.layers.MaxPooling1D(pool_size=conv_layer['max_pool']['pool_size'], strides=conv_layer['max_pool']['strides'] ))
                if conv_layer['batch_normalization'] == True:
                    self.convolutional_block.add(tf.keras.layers.BatchNormalization())

        self.flatten = tf.keras.layers.Flatten()
        
        self.dense_block = tf.keras.models.Sequential(name="dense_block")
        for fc_layer in self.dense_layers:
            self.dense_block.add(tf.keras.layers.Dense(units=fc_layer['n_hidden'], activation=fc_layer['activation']))
            if fc_layer['batch_normalization'] == True:
                self.dense_block.add(tf.keras.layers.BatchNormalization())
            if fc_layer['dropout'] > 0.0:
                self.dense_block.add(tf.keras.layers.Dropout(fc_layer['dropout']))
        self.dense_block.add(tf.keras.layers.Dense(self.n_classes))
        self.dense_block.add(tf.keras.layers.Softmax())
        
    def freeze_conv_block(self):
        """Freeze the convolutional block"""
        self.layers[0].trainable = False

    def unfreeze_conv_block(self):
        """Unfreeze the convolutional block"""
        self.layers[0].trainable = True
    
    def freeze_top(self):
        """Freeze the classification (dense) block"""
        for layer in self.layers[1:]:
            layer.trainable = False
    
    def unfreeze_top(self):
        """Unfreeze the classification (dense) block"""
        for layer in self.layers[1:]:
            layer.trainable = True

    def get_feature_extraction_base(self):
        """ Retrive the feature extraction base (initial convolutional layer + residual blocks)
        
            Returns:
                list containing the feature extraction layers
        """
        return [self.convolutional_block]

    def clone_with_new_top(self, n_classes=None, freeze_base=True):
        """ Clone this instance but replace the original classification top with a new (untrained) one
        
            Args:
                n_classes:int
                    The number of classes the new classification top should output.
                    If None(default), the original number of classes will be used.

                freeze_base:bool
                    If True, the weights of the feature extraction base will be froze (untrainable) in the new model.
                
            Returns:
                cloned_model: instance of CNN1DArch
                    The new model with the old feature extraction base and new classification top.
        """
        if freeze_base == True:
            self.trainable = False

        if n_classes is None:
            n_classes = self.n_classes

        pre_trained_base = self.get_feature_extraction_base()
        cloned_model = type(self)(n_classes=n_classes, pre_trained_base=pre_trained_base)

        return cloned_model


    def call(self, inputs, training=None):
        """Calls the model on new inputs.

        In this case call just reapplies all ops in the graph to the new inputs (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: Tensor or list of tensors
                A tensor or list of tensors
            
            training: Bool
                Boolean or boolean scalar tensor, indicating whether to run the Network in training mode or inference mode.

        Returns:
                A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        output = self.call_frontend(inputs)
        output = self.convolutional_block(output, training=training)
        output = self.flatten(output)
        output = self.dense_block(output, training=training)

        print("output shape: ", output.shape)

        return output


class CNNInterface(NNInterface):
    """ Creates a CNN model with the standardized Ketos interface.

        Args:
            convolutional_layers: list
                A list of dictionaries containing the detailed specification for the convolutional layers.
                Each layer is specified as a dictionary with the following format:

                >>> {'n_filters':96, "filter_shape":(11,11), 'strides':4, 'padding':'valid', activation':'relu', 'max_pool': {'pool_size':(3,3) , 'strides':(2,2)}, 'batch_normalization':True} # doctest: +SKIP

            dense_layers: list
                A list of dictionaries containing the detailed specification for the fully connected layers.
                Each layer is specified as a dictionary with the following format:
                
                >>> {'n_hidden':4096, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5} # doctest: +SKIP

            n_classes:int
                The number of classes the network will be used to classify.
                The output will be this number of values representing the scores for each class. 
                Scores sum to 1.0.

            optimizer: ketos.neural_networks.RecipeCompat object
                A recipe compatible optimizer (i.e.: wrapped by the ketos.neural_networksRecipeCompat class)

            loss_function: ketos.neural_networks.RecipeCompat object
                A recipe compatible loss_function (i.e.: wrapped by the ketos.neural_networksRecipeCompat class)

            metrics: list of ketos.neural_networks.RecipeCompat objects
                A list of recipe compatible metrics (i.e.: wrapped by the ketos.neural_networksRecipeCompat class).
                These metrics will be computed on each batch during training.

           
        Examples:

            >>> # Most users will create a model based on a Ketos recipe 
            >>> # The one below, specifies a CNN with 3 convolutional layers and 2 dense layers
            >>>
            >>> recipe = {'conv_set':[[64, False], [128, True], [256, True]], # doctest: +SKIP
            ...   'dense_set': [512, ],
            ...   'n_classes':2,
            ...   'optimizer': {'recipe_name':'Adam', 'parameters': {'learning_rate':0.005}},
            ...   'loss_function': {'recipe_name':'FScoreLoss', 'parameters':{}},  
            ...   'metrics': [{'recipe_name':'CategoricalAccuracy', 'parameters':{}}]
            ... }
            >>> # To create the CNN, simply  use the  'build_from_recipe' method:
            >>> cnn = CNNInterface._build_from_recipe(recipe, recipe_compat=False) # doctest: +SKIP         
    """
    
    @classmethod
    def _convolutional_layers_from_conv_set(cls, conv_set):
        """ Create a detailed description of the convolutional layers based on the simplified description in 'conv_set'

            The resulting detailed description can then be used to build the convolutional layers in the model

            Args:
                conv_set:list
                    A list describing the convolutional layers in a CNN.
                    each layer is represented by a list of 2 elements: The number of filters (int) and
                    whether or not that layer is followed by a max_pooling operation (boolean).
                    Example:
                   
                    >>> [[64,False], [128, True]] # doctest: +SKIP
                   
                    This conv_set would describe two convolutional layers, with 64 and 128 filters, respectively.
                    Only the second would have max_pooling.

            Returns:
                convolutional_layers: list
                    A list of detailed layer description dictionaries.
                    Example: 
                   
                    >>> [{'n_filters':64, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True}, # doctest: +SKIP
                    ...          {'n_filters':128, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':[2,2] , 'strides':[2,2]}, 'batch_normalization':True},
                    ...          ]
                              

        """
        
        convolutional_layers = []
        for layer_parameters in conv_set:
            n_filters, max_pool = layer_parameters
            
            #default layer details
            layer_details = {'n_filters':64, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':[2,2] , 'strides':[2,2]}, 'batch_normalization':True}
            layer_details['n_filters'] = n_filters
                       
            if max_pool is False:
                layer_details['max_pool'] = None
            convolutional_layers.append(layer_details)
            

        return convolutional_layers


    @classmethod
    def _dense_layers_from_dense_set(cls, dense_set):
        """ Create a detailed description of the dense layers based on the simplified description in 'dense_set'

            The resulting detailed description can then be used to build the convolutional layers in the model.

            Args:
                dense_set:list
                    A list describing the dense layers in a CNN.
                    Each layer is represented by a one integer describing the number of output nodes in that layer.
                    The number of input nodes is automatically determined from the previous layer.
                    Example: [512, 256] 
                    This cdense_set would describe two dense layers, with 512 and 256 nodes, respectively.
                    Note that, the last layer of a CNN does not need to be especified in the dense_set, as the output layer
                    is automatically created according with the number of classes to be classified.

            Returns:
                dense_layers: list
                    A list of detailed layer description dictionaries.
                    Example: 

                    >>> [{'n_hidden':512, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5}, # doctest: +SKIP
                    ...       {'n_hidden':256, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                    ...       ]  

        """

        dense_layers = []
        for layer_parameters in dense_set:
            n_hidden = layer_parameters
            layer_details = {'n_hidden':4096, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5}
            layer_details['n_hidden'] = n_hidden

            dense_layers.append(layer_details)
        return dense_layers


    @classmethod
    def _build_from_recipe(cls, recipe, recipe_compat=True ):
        """ Build a CNN model from a recipe.

            Args:
                recipe: dict
                    A recipe dictionary. The optimizer, loss function
                    and metrics must be instances of ketos.neural_networks.RecipeCompat.
                    Example recipe:

                        >>> {'conv_set':[[64, False], [128, True], [256, True]], # doctest: +SKIP
                        ...  'dense_set': [512, ],
                        ...  'n_classes':2,
                        ...  'optimizer': {'name':'Adam', 'parameters': {'learning_rate':0.005}},
                        ...  'loss_function': {'name':'FScoreLoss', 'parameters':{}},  
                        ...  'metrics': [{'name':'CategoricalAccuracy', 'parameters':{}}]
                        ...  ]

                        The only optional field is 'secondary_metrics'.

                    Alternatively, the 'conv_set' and 'dense_set' can be replaced by detailed descriptions in
                    'convolutional_layers' and 'dense_layers'.
                    Note that these need to be provided in pairs ('conv_set' + 'dense_set' OR 'convolutional_layers' and 'dense_layers')
                    Example:

                        >>> {'conv_set': [{'n_filters':64, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True}, # doctest: +SKIP
                        ...              {'n_filters':128, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':[2,2] , 'strides':[2,2]}, 'batch_normalization':True},
                        ...              ],
                        ...  'dense_set': [{'n_hidden':512, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                        ...                {'n_hidden':256, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                        ...                ],
                        ...  'n_classes':2,
                        ...  'optimizer': {'name':'Adam', 'parameters': {'learning_rate':0.005}},
                        ...  'loss_function': {'name':'FScoreLoss', 'parameters':{}},  
                        ...  'metrics': [{'name':'CategoricalAccuracy', 'parameters':{}}],
                        ...  ]


            Returns:
                An instance of CNNInterface.
        """

        conv_set = None
        dense_set = None        
        if 'convolutional_layers' in recipe.keys() and 'dense_layers' in recipe.keys():
            convolutional_layers = recipe['convolutional_layers']
            dense_layers = recipe['dense_layers']
        elif 'conv_set' in recipe.keys() and 'dense_set' in recipe.keys():
            conv_set = recipe['conv_set']
            dense_set = recipe['dense_set']
            convolutional_layers = cls._convolutional_layers_from_conv_set(conv_set)
            dense_layers = cls._dense_layers_from_dense_set(dense_set)
            
        n_classes = recipe['n_classes']
        
        if recipe_compat == True:
            optimizer = recipe['optimizer']
            loss_function = recipe['loss_function']
            metrics = recipe['metrics']
            
        else:
            optimizer = cls._optimizer_from_recipe(recipe['optimizer'])
            loss_function = cls._loss_function_from_recipe(recipe['loss_function'])
            metrics = cls._metrics_from_recipe(recipe['metrics'])
            
        

        instance = cls(convolutional_layers=convolutional_layers, dense_layers=dense_layers, n_classes=n_classes, optimizer=optimizer, loss_function=loss_function, metrics=metrics)
        instance.conv_set = conv_set
        instance.dense_set = dense_set

        return instance
   
    @classmethod
    def _read_recipe_file(cls, json_file, return_recipe_compat=True):
        """ Read a CNN recipe saved in a .json file.

            Args:
                json_file:string
                    Full path (including silename and extension) to the .json file containing the recipe.
                return_recipe_compat:bool
                    If True, returns a dictionary where the optimizer, loss_function, metrics and 
                    secondary_metrics (if available) values are instances of the ketos.neural_networks.nn_interface.RecipeCompat.
                        The returned dictionary will be equivalent to:

                           >>> {'conv_set':[(64, False), (128, True), (256, True)], # doctest: +SKIP
                           ...  'dense_set: [512, 256],
                           ...  'convolutional_layers: ,
                           ...  'dense_layers: ,
                           ... 'n_classes': 2 ,
                           ... 'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                           ... 'loss_function': RecipeCompat('FScoreLoss', FScoreLoss),  
                           ... 'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)]}

                    If False, the optimizer, loss_function, metrics and secondary_metrics (if available) values will contain a
                    dictionary representation of such fields instead of the RecipeCompat objects:
                        
                            >>> {'conv_set':[(64, False), (128, True), (256, True)], # doctest: +SKIP
                            ...    'dense_set: [512, 256],
                            ...    'convolutional_layers: ,
                            ...    'dense_layers: ,
                            ...    'n_classes': 2 ,
                            ...    'optimizer': {'name':'Adam', 'parameters': {'learning_rate':0.005}},
                            ...    'loss_function': {'name':'FScoreLoss', 'parameters':{}},  
                            ...    'metrics': [{'name':'CategoricalAccuracy', 'parameters':{}}]}

                Returns:
                    recipe, according to 'return_recipe_compat'.

        """

        with open(json_file, 'r') as json_recipe:
            recipe_dict = json.load(json_recipe)

        optimizer = cls._optimizer_from_recipe(recipe_dict['optimizer'])
        loss_function = cls._loss_function_from_recipe(recipe_dict['loss_function'])
        metrics = cls._metrics_from_recipe(recipe_dict['metrics'])
        

        if return_recipe_compat == True:
            recipe_dict['optimizer'] = optimizer
            recipe_dict['loss_function'] = loss_function
            recipe_dict['metrics'] = metrics
            
        else:
            recipe_dict['optimizer'] = cls._optimizer_to_recipe(optimizer)
            recipe_dict['loss_function'] = cls._loss_function_to_recipe(loss_function)
            recipe_dict['metrics'] = cls._metrics_to_recipe(metrics)
            
        if 'convolutional_layers' in recipe_dict.keys() and 'dense_layers' in recipe_dict.keys():
            convolutional_layers = recipe_dict['convolutional_layers']
            dense_layers = recipe_dict['dense_layers']
        elif 'conv_set' in recipe.keys() and 'dense_set' in recipe_dict.keys():
            conv_set = recipe_dict['conv_set']
            dense_set = recipe_dict['dense_set']
            convolutional_layers = cls._convolutional_layers_from_conv_set(conv_set)
            dense_layers = cls._dense_layers_from_dense_set(dense_set)
            
        recipe_dict['conv_set'] = recipe_dict['conv_set']
        recipe_dict['dense_set'] = recipe_dict['dense_set']
        recipe_dict['convolutional_layers'] = recipe_dict['convolutional_layers']
        recipe_dict['dense_layers'] = recipe_dict['dense_layers']
        recipe_dict['n_classes'] = recipe_dict['n_classes']
        

        return recipe_dict


    def __init__(self, convolutional_layers=default_cnn_recipe['convolutional_layers'], dense_layers=default_cnn_recipe['dense_layers'],
                 n_classes=default_cnn_recipe['n_classes'], optimizer=default_cnn_recipe['optimizer'], loss_function=default_cnn_recipe['loss_function'], 
                 metrics=default_cnn_recipe['metrics']):
        super(CNNInterface, self).__init__(optimizer, loss_function, metrics)
        self.conv_set = None
        self.dense_det = None
        self.convolutional_layers = convolutional_layers
        self.dense_layers = dense_layers
        self.n_classes = n_classes
       
        self.model=CNNArch(convolutional_layers=self.convolutional_layers, dense_layers=self.dense_layers, n_classes=n_classes)
       


    def _extract_recipe_dict(self):
        """ Create a recipe dictionary from a CNNInterface instance.

            The resulting recipe contains all the fields necessary to build the same network architecture used by the instance calling this method.
            
            Returns:
                recipe:dict
                    A dictionary containing the recipe fields necessary to build the same network architecture.
                    Example:

                        >>> {'conv_set':[(64, False), (128, True), (256, True)], # doctest: +SKIP
                        ...  'dense_set: [512, 256],
                        ...  'convolutional_layers: ,
                        ...  'dense_layers: ,
                        ...  'n_classes':2,
                        ...  'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                        ...  'loss_function': RecipeCompat('FScoreLoss', FScoreLoss),  
                        ...  'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)]}
        """

        recipe = {}
        recipe['interface'] = type(self).__name__
        recipe['conv_set'] = self.conv_set
        recipe['dense_set'] = self.dense_set
        recipe['convolutional_layers'] = self.convolutional_layers
        recipe['dense_layers'] = self.dense_layers
        recipe['n_classes'] = self.n_classes
        recipe['optimizer'] = self._optimizer_to_recipe(self.optimizer)
        recipe['loss_function'] = self._loss_function_to_recipe(self.loss_function)
        recipe['metrics'] = self._metrics_to_recipe(self.metrics)
        
        return recipe


class CNN1DInterface(CNNInterface):
    """ Create an 1D (temporal) CNN model with the standardized Ketos interface.

        Args:
            convolutional_layers: list
                A list of dictionaries containing the detailed specification for the convolutional layers.
                Each layer is specified as a dictionary with the following format:

                >>> {'n_filters':96, "filter_shape":(11,11), 'strides':4, 'padding':'valid', activation':'relu', 'max_pool': {'pool_size':(3,3) , 'strides':(2,2)}, 'batch_normalization':True} # doctest: +SKIP

            dense_layers: list
                A list of dictionaries containing the detailed specification for the fully connected layers.
                Each layer is specified as a dictionary with the following format:

                >>> {'n_hidden':4096, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5} # doctest: +SKIP

            n_classes:int
                The number of classes the network will be used to classify.
                The output will be this number of values representing the scores for each class. 
                Scores sum to 1.0.

            optimizer: ketos.neural_networks.RecipeCompat object
                A recipe compatible optimizer (i.e.: wrapped by the ketos.neural_networksRecipeCompat class)

            loss_function: ketos.neural_networks.RecipeCompat object
                A recipe compatible loss_function (i.e.: wrapped by the ketos.neural_networksRecipeCompat class)

            metrics: list of ketos.neural_networks.RecipeCompat objects
                A list of recipe compatible metrics (i.e.: wrapped by the ketos.neural_networksRecipeCompat class).
                These metrics will be computed on each batch during training.

         
        Examples:

            >>> # Most users will create a model based on a Ketos recipe 
            >>> # The one below, specifies a CNN with 3 convolutional layers and 2 dense layers
            >>>
            >>> recipe = {'conv_set':[[64, False], [128, True], [256, True]], # doctest: +SKIP
            ...   'dense_set': [512, ],
            ...   'n_classes':2,
            ...   'optimizer': {'name':'Adam', 'parameters': {'learning_rate':0.005}},
            ...   'loss_function': {'name':'FScoreLoss', 'parameters':{}},  
            ...   'metrics': [{'name':'CategoricalAccuracy', 'parameters':{}}]
            ... }
            >>> # To create the CNN, simply  use the  'build_from_recipe' method:
            >>> cnn = CNNInterface.build_from_recipe(recipe, recipe_compat=False) # doctest: +SKIP
                
    """

    @classmethod
    def _convolutional_layers_from_conv_set(cls, conv_set):
        """ Create a detailed description of the convolutional layers based on the simplified description in 'conv_set'

            The resulting detailed description can then be used to build the convolutional layers in the model

            Args:
                conv_set:list
                    A list describing the convolutional layers in a CNN.
                    each layer is represented by a list of 2 elements: The number of filters (int) and
                    whether or not that layer is followed by a max_pooling operation (boolean).
                    Example:
                   
                    >>> [[64,False], [128, True]] # doctest: +SKIP
                   
                    This conv_set would describe two convolutional layers, with 64 and 128 filters, respectively.
                    Only the second would have max_pooling.

            Returns:
                convolutional_layers: list
                    A list of detailed layer description dictionaries.
                    Example: 
                   
                    >>> [{'n_filters':64, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True}, # doctest: +SKIP
                    ...          {'n_filters':128, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':[2,2] , 'strides':[2,2]}, 'batch_normalization':True},
                    ...          ]
                              

        """
        
        convolutional_layers = []
        for layer_parameters in conv_set:
            n_filters, max_pool = layer_parameters
            
            #default layer details
            layer_details = {'n_filters':64, "filter_shape":64, 'strides':2, 'padding':'causal', 'activation':'relu', 'max_pool':{'pool_size':8 , 'strides':8}, 'batch_normalization':True}
            layer_details['n_filters'] = n_filters
                       
            if max_pool is False:
                layer_details['max_pool'] = None
            convolutional_layers.append(layer_details)
            

        return convolutional_layers


    @classmethod
    def _dense_layers_from_dense_set(cls, dense_set):
        """ Create a detailed description of the dense layers based on the simplified description in 'dense_set'

            The resulting detailed description can then be used to build the convolutional layers in the model.

            Args:
                dense_set:list
                    A list describing the dense layers in a CNN.
                    Each layer is represented by a one integer describing the number of output nodes in that layer.
                    The number of input nodes is automatically determined from the previous layer.
                    Example: [512, 256] 
                    This cdense_set would describe two dense layers, with 512 and 256 nodes, respectively.
                    Note that, the last layer of a CNN does not need to be especified in the dense_set, as the output layer
                    is automatically created according with the number of classes to be classified.

            Returns:
                dense_layers: list
                    A list of detailed layer description dictionaries.
                    Example: 

                    >>> [{'n_hidden':512, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5}, # doctest: +SKIP
                    ...       {'n_hidden':256, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                    ...       ]  

        """

        dense_layers = []
        for layer_parameters in dense_set:
            n_hidden = layer_parameters
            layer_details = {'n_hidden':512, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5}
            layer_details['n_hidden'] = n_hidden

            dense_layers.append(layer_details)
        return dense_layers


    @classmethod
    def _build_from_recipe(cls, recipe, recipe_compat=True ):
        """ Build a CNN model from a recipe.

            Args:
                recipe: dict
                    A recipe dictionary. The optimizer, loss function
                    and metrics must be instances of ketos.neural_networks.RecipeCompat.
                    Example recipe:

                        >>> {'conv_set':[[64, False], [128, True], [256, True]], # doctest: +SKIP
                        ...  'dense_set': [512, ],
                        ...  'n_classes':2,
                        ...  'optimizer': {'name':'Adam', 'parameters': {'learning_rate':0.005}},
                        ...  'loss_function': {'name':'FScoreLoss', 'parameters':{}},  
                        ...  'metrics': [{'name':'CategoricalAccuracy', 'parameters':{}}]
                        ...  ]

                        The only optional field is 'secondary_metrics'.

                    Alternatively, the 'conv_set' and 'dense_set' can be replaced by detailed descriptions in
                    'convolutional_layers' and 'dense_layers'.
                    Note that these need to be provided in pairs ('conv_set' + 'dense_set' OR 'convolutional_layers' and 'dense_layers')
                    Example:

                        >>> {'conv_set': [{'n_filters':64, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True}, # doctest: +SKIP
                        ...              {'n_filters':128, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':[2,2] , 'strides':[2,2]}, 'batch_normalization':True},
                        ...              ],
                        ...  'dense_set': [{'n_hidden':512, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                        ...                {'n_hidden':256, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                        ...                ],
                        ...  'n_classes':2,
                        ...  'optimizer': {'name':'Adam', 'parameters': {'learning_rate':0.005}},
                        ...  'loss_function': {'name':'FScoreLoss', 'parameters':{}},  
                        ...  'metrics': [{'name':'CategoricalAccuracy', 'parameters':{}}],
                        ...  ]


            Returns:
                An instance of CNNInterface.
        """

        conv_set = None
        dense_set = None        
        if 'convolutional_layers' in recipe.keys() and 'dense_layers' in recipe.keys():
            convolutional_layers = recipe['convolutional_layers']
            dense_layers = recipe['dense_layers']
        elif 'conv_set' in recipe.keys() and 'dense_set' in recipe.keys():
            conv_set = recipe['conv_set']
            dense_set = recipe['dense_set']
            convolutional_layers = cls._convolutional_layers_from_conv_set(conv_set)
            dense_layers = cls._dense_layers_from_dense_set(dense_set)
            
        n_classes = recipe['n_classes']
        
        if recipe_compat == True:
            optimizer = recipe['optimizer']
            loss_function = recipe['loss_function']
            metrics = recipe['metrics']
            
        else:
            optimizer = cls._optimizer_from_recipe(recipe['optimizer'])
            loss_function = cls._loss_function_from_recipe(recipe['loss_function'])
            metrics = cls._metrics_from_recipe(recipe['metrics'])
            
        

        instance = cls(convolutional_layers=convolutional_layers, dense_layers=dense_layers, n_classes=n_classes, optimizer=optimizer, loss_function=loss_function, metrics=metrics)
        instance.conv_set = conv_set
        instance.dense_set = dense_set

        return instance

    @classmethod
    def transform_batch(cls, x, y, n_classes=2):
        """ Transforms a training batch into the format expected by the network.

            When this interface is subclassed to make new neural_network classes, this method can be overwritten to
            accomodate any transformations required. Common operations are reshaping of input arrays and parsing or one hot encoding of the labels.

            Args:
                x:numpy.array
                    The batch of inputs with shape (batch_size, width, height)
                y:numpy.array
                    The batch of labels.
                    Each label must be represented as an integer, ranging from zero to n_classes
                    The array is expected to have a field named 'label'.
                n_classes:int
                    The number of possible classes for one hot encoding.
                    
                

            Returns:
                X:numpy.array
                    The transformed batch of inputs
                Y:numpy.array
                    The transformed batch of labels

            Examples:

                >>> import numpy as np
                >>> # Create a batch of 10 5x5 arrays
                >>> inputs = np.random.rand(10,5,5)
                >>> inputs.shape
                (10, 5, 5)

                    
                >>> # Create a batch of 10 labels (0 or 1)
                >>> labels = np.random.choice([0,1], size=10)
                >>> labels.shape
                (10,)

                >>> transformed_inputs, transformed_labels = NNInterface.transform_batch(inputs, labels, n_classes=2)
                >>> transformed_inputs.shape
                (10, 5, 5, 1)

                >>> transformed_labels.shape
                (10, 2)
                
        """

        X = cls._transform_input(x)
        if y.dtype.names is not None:
            Y = np.array([cls._to1hot(class_label=label, n_classes=n_classes) for label in y['label']])
        else:
            Y = np.array([cls._to1hot(class_label=label, n_classes=n_classes) for label in y])

        return (X,Y)

    @classmethod
    def _transform_input(cls,input):
        """ Transforms a training input to the format expected by the network.

            Similar to :func:`NNInterface.transform_train_batch`, but only acts on the inputs (not labels). Mostly used for inference, rather than training.
            When this interface is subclassed to make new neural_network classes, this method can be overwritten to
            accomodate any transformations required. Common operations are reshaping of an input.

            Args:
                input:numpy.array
                    An input instance. Must be of shape (n,m) or (k,n,m).

            Raises:
                ValueError if input does not have 2 or 3 dimensions.

            Returns:
                tranformed_input:numpy.array
                    The transformed batch of inputs

            Examples:
            
                >>> import numpy as np
                >>> # Create a batch of 10 5x5 arrays
                >>> batch_of_inputs = np.random.rand(10,5,5)
                >>> selected_input = batch_of_inputs[0]
                >>> selected_input.shape
                (5, 5)
                 
                >>> transformed_input = NNInterface._transform_input(selected_input)
                >>> transformed_input.shape
                (1, 5, 5, 1)

                # The input can also have shape=(1,n,m)
                >>> selected_input = batch_of_inputs[0:1]
                >>> selected_input.shape
                (1, 5, 5)
                 
                >>> transformed_input = NNInterface._transform_input(selected_input)
                >>> transformed_input.shape
                (1, 5, 5, 1)

                
        """

        if input.ndim == 1:
            transformed_input = input.reshape(1,input.shape[0],1)
        elif input.ndim == 2:
            transformed_input = input.reshape(input.shape[0],input.shape[1],1)
        else:
            raise ValueError("Expected input to have 1 or 2 dimensions, got {}({}) instead".format(input.ndims, input.shape))

        return transformed_input

    def __init__(self, convolutional_layers=default_cnn_1d_recipe['convolutional_layers'], dense_layers=default_cnn_1d_recipe['dense_layers'],
                 n_classes=default_cnn_1d_recipe['n_classes'], optimizer=default_cnn_1d_recipe['optimizer'], loss_function=default_cnn_1d_recipe['loss_function'], 
                 metrics=default_cnn_1d_recipe['metrics']):
        super(CNN1DInterface, self).__init__(optimizer=optimizer, loss_function=loss_function, metrics=metrics)
        self.conv_set = None
        self.dense_set = None
        self.convolutional_layers = convolutional_layers
        self.dense_layers = dense_layers
        self.n_classes = n_classes
       
        self.model = CNN1DArch(convolutional_layers=self.convolutional_layers, dense_layers=self.dense_layers, n_classes=n_classes)
       



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" resnet sub-module within the ketos.neural_networks module

    This module provides classes to implement Residual Networks (ResNets).

    Contents:
        ResNetBlock class
        ResNet class
        ResNetInterface class

"""

import tensorflow as tf
import numpy as np
from ketos.neural_networks.dev_utils.nn_interface import RecipeCompat, NNInterface, NNArch
import json




default_resnet_recipe =  {'block_sets':[2,2,2],
                    'n_classes':2,
                    'initial_filters':16,        
                    'initial_strides':1,
                    'initial_kernel':[3,3],        
                    'strides':2,
                    'kernel':[3,3],        
                    'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                    'loss_function': RecipeCompat('BinaryCrossentropy', tf.keras.losses.BinaryCrossentropy),  
                    'metrics': [RecipeCompat('BinaryAccuracy',tf.keras.metrics.BinaryAccuracy),
                                RecipeCompat('Precision',tf.keras.metrics.Precision),
                                RecipeCompat('Recall',tf.keras.metrics.Recall)],
                    }


default_resnet_1d_recipe =  {'block_sets':[2,2,2],
                    'n_classes':2,
                    'initial_filters':2,
                    'initial_strides':1,
                    'initial_kernel':30,        
                    'strides':2,
                    'kernel':300,        
                    'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                    'loss_function': RecipeCompat('CategoricalCrossentropy', tf.keras.losses.CategoricalCrossentropy),  
                    'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy),
                                RecipeCompat('Precision',tf.keras.metrics.Precision, class_id=1),
                                RecipeCompat('Recall',tf.keras.metrics.Recall, class_id=1)],
                    }



class ResNetBlock(tf.keras.Model):
    """ Residual block for ResNet architectures.

        Args: 
            filters: int
                The number of filters in the block
            strides: int
                Strides used in convolutional layers within the block
            kernel: (int,int)
                Kernel used in convolutional layers within the block
            residual_path: bool
                Whether or not the block will contain a residual path
            batch_norm_momentum: float between 0 and 1
                Momentum for the moving average of the batch normalization layers.
                The default value is 0.99.
                For an explanation of how the momentum affects the batch normalisation operation,
                see <https://www.tensorflow.org/api_docs/python/tf/keras/layers/BatchNormalization>
            dropout_rate: float between 0 and 1
                Fraction of the input units to drop in the dropout layers.
                Set this parameter to 0 to disable dropout (default).
        Returns:
            A ResNetBlock object. The block itself is a tensorflow model and can be used as such.
    """
    def __init__(self, filters, strides=1, kernel=(3,3), residual_path=False, batch_norm_momentum=0.99, dropout_rate=0):
        super(ResNetBlock, self).__init__()

        self.filters = filters
        self.strides = strides
        self.kernel  = kernel
        self.residual_path = residual_path

        self.conv_1 = tf.keras.layers.Conv2D(filters=self.filters, kernel_size=self.kernel, strides=self.strides,
                                                padding="same", use_bias=False,
                                                kernel_initializer=tf.random_normal_initializer())

        self.batch_norm_1 = tf.keras.layers.BatchNormalization(momentum=batch_norm_momentum)

        self.conv_2 = tf.keras.layers.Conv2D(filters=self.filters, kernel_size=self.kernel, strides=1,
                                                padding="same", use_bias=False,
                                                kernel_initializer=tf.random_normal_initializer())

        self.batch_norm_2 = tf.keras.layers.BatchNormalization(momentum=batch_norm_momentum)

        if residual_path == True:
            self.conv_down = tf.keras.layers.Conv2D(filters=self.filters, kernel_size=(1,1), strides=self.strides,
                                                padding="same", use_bias=False,
                                                kernel_initializer=tf.random_normal_initializer())

            self.batch_norm_down = tf.keras.layers.BatchNormalization(momentum=batch_norm_momentum)
        
        self.dropout = tf.keras.layers.Dropout(rate=dropout_rate)

    def set_batch_norm_momentum(self, momentum):
        """ Set the momentum for the moving average of the batch normalization layers in the block.

            For an explanation of how the momentum affects the batch normalisation operation,
            see <https://www.tensorflow.org/api_docs/python/tf/keras/layers/BatchNormalization>

            Args: 
                momentum: float between 0 and 1
                    Momentum for the moving average of the batch normalization layers.

            Returns:
                None
        """
        assert momentum>=0 and momentum<=1, 'batch normalization momentum must be between 0 and 1'
        
        self.batch_norm_1.momentum = momentum
        self.batch_norm_2.momentum = momentum
        if self.residual_path:
            self.batch_norm_down.momentum = momentum

    def set_dropout_rate(self, rate):
        """ Set the fraction of the input units to drop in the dropout layers in the block.

            Args: 
                rate: float between 0 and 1
                    Fraction of the input units to drop in the dropout layers.

            Returns:
                None
        """
        assert rate>=0 and rate<=1, 'dropout rate must be between 0 and 1'

        self.dropout.rate = rate

    def call(self,inputs, training=None):
        """Calls the model on new inputs.

        In this case call just reapplies all ops in the graph to the new inputs (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: Tensor or list of tensors
                A tensor or list of tensors
            
            training: Bool
                Boolean or boolean scalar tensor, indicating whether to run the Network in training mode or inference mode.

        Returns:
                A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        residual = inputs

        x = self.batch_norm_1(inputs, training=training)
        x = tf.nn.relu(x)
        x = self.conv_1(x)
        x = self.dropout(x, training=training)
        x = self.batch_norm_2(x, training=training)
        x = tf.nn.relu(x)
        x = self.conv_2(x)
        x = self.dropout(x, training=training)

        if self.residual_path:
            residual = self.batch_norm_down(inputs, training=training)
            residual = tf.nn.relu(residual)
            residual = self.conv_down(residual)
            x = self.dropout(x, training=training)

        x = x + residual
        return x


class ResNet1DBlock(tf.keras.Model):
    """ Residual block for 1D (temporal) ResNet architectures.

        Args: 
            filters: int
                The number of filters in the block
            strides: int
                Strides used in convolutional layers within the block
            kernel: int
                Kernel size used in convolutional layers within the block
            residual_path: bool
                Whether or not the block will contain a residual path
            batch_norm_momentum: float between 0 and 1
                Momentum for the moving average of the batch normalization layers.
                The default value is 0.99.
                For an explanation of how the momentum affects the batch normalisation operation,
                see <https://www.tensorflow.org/api_docs/python/tf/keras/layers/BatchNormalization>
            dropout_rate: float between 0 and 1
                Fraction of the input units to drop in the dropout layers.
                Set this parameter to 0 to disable dropout (default).

        Returns:
            A ResNetBlock object. The block itself is a tensorflow model and can be used as such.
    """
    def __init__(self, filters, strides=1, kernel=300, residual_path=False, batch_norm_momentum=0.99, dropout_rate=0):
        super(ResNet1DBlock, self).__init__()

        self.filters = filters
        self.strides = strides
        self.kernel = kernel
        self.residual_path = residual_path

        self.conv_1 = tf.keras.layers.Conv1D(filters=self.filters, kernel_size=self.kernel, strides=self.strides,
                                                padding="same", use_bias=False,
                                                kernel_initializer=tf.random_normal_initializer())

        self.batch_norm_1 = tf.keras.layers.BatchNormalization(momentum=batch_norm_momentum)

        self.conv_2 = tf.keras.layers.Conv1D(filters=self.filters, kernel_size=self.kernel, strides=1,
                                                padding="same", use_bias=False,
                                                kernel_initializer=tf.random_normal_initializer())

        self.batch_norm_2 = tf.keras.layers.BatchNormalization(momentum=batch_norm_momentum)

        if residual_path == True:
            self.conv_down = tf.keras.layers.Conv1D(filters=self.filters, kernel_size=1, strides=self.strides,
                                                padding="same", use_bias=False,
                                                kernel_initializer=tf.random_normal_initializer())

            self.batch_norm_down = tf.keras.layers.BatchNormalization(momentum=batch_norm_momentum)
        
        self.dropout = tf.keras.layers.Dropout(rate=dropout_rate)


    def set_batch_norm_momentum(self, momentum):
        """ Set the momentum for the moving average of the batch normalization layers in the block.

            For an explanation of how the momentum affects the batch normalisation operation,
            see <https://www.tensorflow.org/api_docs/python/tf/keras/layers/BatchNormalization>

            Args: 
                momentum: float between 0 and 1
                    Momentum for the moving average of the batch normalization layers.

            Returns:
                None
        """
        assert momentum>=0 and momentum<=1, 'batch normalization momentum must be between 0 and 1'
        
        self.batch_norm_1.momentum = momentum
        self.batch_norm_2.momentum = momentum
        if self.residual_path:
            self.batch_norm_down.momentum = momentum

    def set_dropout_rate(self, rate):
        """ Set the fraction of the input units to drop in the dropout layers in the block.

            Args: 
                rate: float between 0 and 1
                    Fraction of the input units to drop in the dropout layers.

            Returns:
                None
        """
        assert rate>=0 and rate<=1, 'dropout rate must be between 0 and 1'

        self.dropout.rate = rate

    def call(self,inputs, training=None):
        """Calls the model on new inputs.

        In this case call just reapplies all ops in the graph to the new inputs (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: Tensor or list of tensors
                A tensor or list of tensors
            
            training: Bool
                Boolean or boolean scalar tensor, indicating whether to run the Network in training mode or inference mode.

        Returns:
                A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        residual = inputs

        x = self.batch_norm_1(inputs, training=training)
        x = tf.nn.relu(x)
        x = self.conv_1(x)
        x = self.dropout(x, training=training)
        x = self.batch_norm_2(x, training=training)
        x = tf.nn.relu(x)
        x = self.conv_2(x)
        x = self.dropout(x, training=training)

        if self.residual_path:
            residual = self.batch_norm_down(inputs, training=training)
            residual = tf.nn.relu(residual)
            residual = self.conv_down(residual)
            x = self.dropout(x, training=training)

        x = x + residual
        return x


class ResNetArch(NNArch):
    """ Implements a ResNet architecture, building on top of ResNetBlocks.

        Args:
            block_sets: list of ints
                A list specifying the block sets and how many blocks each  set contains.
                Example: [2,2,2] will create a ResNet with 3 block sets, each containing
                2 ResNetBlocks (i.e.: a total of 6 residual blocks)
            
            n_classes:int
                The number of classes. The output layer uses a Softmax activation and
                will contain this number of nodes, resulting in model outputs with this
                many values summing to 1.0.

            initial_filters:int
                The number of filters used in the first ResNetBlock. Subsequent blocks 
                will have two times more filters than their previous block.

            initial_strides: int
                Strides used in the first convolutional layer

            initial_kernel: (int,int)
                Kernel used in the first convolutional layer

            strides: int
                Strides used in convolutional layers within the block

            kernel: (int,int)
                Kernel used in convolutional layers within the block

            pre_trained_base: instance of ResNetArch
                A pre-trained resnet model from which the residual blocks will be taken. 
                Use by the the clone_with_new_top method when creating a clone for transfer learning

            batch_norm_momentum: float between 0 and 1
                Momentum for the moving average of all the batch normalization layers in the network.
                The default value is 0.99.
                For an explanation of how the momentum affects the batch normalisation operation,
                see <https://www.tensorflow.org/api_docs/python/tf/keras/layers/BatchNormalization>

            dropout_rate: float between 0 and 1
                Fraction of the input units to drop in all the dropout layers in the network.
                Set this parameter to 0 to disable dropout (default).

        Returns:
            A ResNetArch object, which is a tensorflow model.
    """

    def __init__(self,  n_classes, pre_trained_base=None, block_sets=None, initial_filters=16, 
                        initial_strides=1, initial_kernel=(3,3), strides=2, kernel=(3,3), 
                        batch_norm_momentum=0.99, dropout_rate=0, **kwargs):
        super(ResNetArch, self).__init__(**kwargs)

        self.n_classes = n_classes
        self.initial_strides = initial_strides
        self.initial_kernel = initial_kernel
        self.strides = strides
        self.kernel = kernel

        if pre_trained_base:
            self.conv_initial = pre_trained_base[0]
            self.blocks = pre_trained_base[1]

        else:
            self.n_sets = len(block_sets)
            self.block_sets = block_sets
            self.input_filters = initial_filters
            self.output_filters = initial_filters
            self.conv_initial = tf.keras.layers.Conv2D(filters=self.output_filters, strides=self.initial_strides, 
                                                    kernel_size=self.initial_kernel, padding="same", use_bias=False,
                                                    kernel_initializer=tf.random_normal_initializer())

            self.blocks = tf.keras.models.Sequential(name="dynamic_blocks")

            self.num_blocks = 0
            for set_id in range(self.n_sets):
                for block_id in range(self.block_sets[set_id]):
                    #First layer of every block except the first
                    if set_id != 0 and block_id == 0:
                        block = ResNetBlock(self.output_filters, strides=self.strides, kernel=self.kernel, residual_path=True,
                                            batch_norm_momentum=batch_norm_momentum, dropout_rate=dropout_rate)
                    
                    else:
                        if self.input_filters != self.output_filters:
                            residual_path = True
                        else:
                            residual_path = False

                        block = ResNetBlock(self.output_filters, strides=1, kernel=self.kernel, residual_path=residual_path,
                                            batch_norm_momentum=batch_norm_momentum, dropout_rate=dropout_rate)

                    self.input_filters = self.output_filters

                    self.blocks.add(block)
                    self.num_blocks += 1
                
                self.output_filters *= 2

        self.batch_norm_final = tf.keras.layers.BatchNormalization(momentum=batch_norm_momentum)
        self.average_pool = tf.keras.layers.GlobalAveragePooling2D()
        self.fully_connected = tf.keras.layers.Dense(self.n_classes)
        self.softmax = tf.keras.layers.Softmax()
    
    def freeze_init_layer(self):
        """Freeze the initial convolutional layer"""
        self.layers[0].trainable = False

    def unfreeze_init_layer(self):
        """Unffreeze the initial convolutional layer"""
        self.layers[0].trainable = True
    
    def freeze_block(self, block_ids):
        """ Freeze specific residual blocks

            Args:
                blocks_ids: int
                    The block number to be freezed (starting from zero)
        """

        for block_id in block_ids:
            self.layers[1].layers[block_id].trainable = False

    def unfreeze_block(self, block_ids):
        """ Unfreeze specific residual blocks

            Args:
                blocks_ids: int
                    The block number to be unfreezed (starting from zero)
        """

        for block_id in block_ids:
            self.layers[1].layers[block_id].trainable = True
    
    def freeze_top(self):
        """Freeze the classification block"""
        for layer in self.layers[2:]:
            layer.trainable = False
    
    def unfreeze_top(self):
        """Unfreeze the classification block"""
        for layer in self.layers[2:]:
            layer.trainable = True

    def get_feature_extraction_base(self):
        """ Retrive the feature extraction base (initial convolutional layer + residual blocks)
        
            Returns:
                list containing the feature extraction layers
        """
        return [self.conv_initial, self.blocks]

    def clone_with_new_top(self, n_classes=None, freeze_base=True):
        """ Clone this instance but replace the original classification top with a new (untrained) one
        
            Args:
                n_classes:int
                    The number of classes the new classification top should output.
                    If None(default), the original number of classes will be used.
                freeze_base:bool
                    If True, the weights of the feature extraction base will be froze (untrainable) in the new model.
                
            Returns:
                cloned_model: instance of ResNetArch
                    The new model with the old feature extraction base and new classification top.
         """
        if freeze_base == True:
            self.trainable = False

        if n_classes is None:
            n_classes = self.n_classes

        pre_trained_base = self.get_feature_extraction_base()
        cloned_model = type(self)(n_classes=n_classes, pre_trained_base=pre_trained_base)

        return cloned_model

    def set_batch_norm_momentum(self, momentum):
        """ Set the momentum for the moving average of all the batch normalization layers in the network.

            For an explanation of how the momentum affects the batch normalisation operation,
            see <https://www.tensorflow.org/api_docs/python/tf/keras/layers/BatchNormalization>

            Args: 
                momentum: float between 0 and 1
                    Momentum for the moving average of the batch normalization layers.

            Returns:
                None
        """
        assert momentum>=0 and momentum<=1, 'batch normalization momentum must be between 0 and 1'

        self.batch_norm_final.momentum = momentum
        for block_id in range(self.num_blocks):
            self.layers[1].layers[block_id].set_batch_norm_momentum(momentum=momentum)

    def set_dropout_rate(self, rate):
        """ Set the fraction of the input units to drop in all the dropout layers in the network.

            Args: 
                rate: float between 0 and 1
                    Fraction of the input units to drop in the dropout layers.

            Returns:
                None
        """
        assert rate>=0 and rate<=1, 'dropout rate must be between 0 and 1'

        for block_id in range(self.num_blocks):
            self.layers[1].layers[block_id].set_dropout_rate(rate=rate)

    def call(self, inputs, training=None):
        """Calls the model on new inputs.

        In this case call just reapplies all ops in the graph to the new inputs (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: Tensor or list of tensors
                A tensor or list of tensors
            
            training: Bool
                Boolean or boolean scalar tensor, indicating whether to run the Network in training mode or inference mode.

        Returns:
                A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        output = self.call_frontend(inputs)
        output = self.conv_initial(output)
        output = self.blocks(output, training=training)
        output = self.batch_norm_final(output, training=training)
        output = tf.nn.relu(output)
        output = self.average_pool(output)
        output = self.fully_connected(output)
        output = self.softmax(output)

        return output
    


class ResNet1DArch(NNArch):
    """ Implements a 1D (temporal) ResNet architecture, building on top of ResNetBlocks.

        Args:
            block_sets: list of ints
                A list specifying the block sets and how many blocks each  set contains.
                Example: [2,2,2] will create a ResNet with 3 block sets, each containing
                2 ResNetBlocks (i.e.: a total of 6 residual blocks)
            
            n_classes:int
                The number of classes. The output layer uses a Softmax activation and
                will contain this number of nodes, resulting in model outputs with this
                many values summing to 1.0.

            initial_filters:int
                The number of filters used in the first ResNetBlock. Subsequent blocks 
                will have two times more filters than their previous block.

            initial_strides: int
                Strides used in the first convolutional layer

            initial_kernel: int
                Kernel size used in the first convolutional layer

            strides: int
                Strides used in convolutional layers within the blocks

            kernel: int
                Kernel size used in convolutional layers within the blocks

            pre_trained_base: instance of ResNet1DArch
                A pre-trained resnet model from which the residual blocks will be taken. 
                Use by the the clone_with_new_top method when creating a clone for transfer learning

            batch_norm_momentum: float between 0 and 1
                Momentum for the moving average of all the batch normalization layers in the network.
                The default value is 0.99.
                For an explanation of how the momentum affects the batch normalisation operation,
                see <https://www.tensorflow.org/api_docs/python/tf/keras/layers/BatchNormalization>

            dropout_rate: float between 0 and 1
                Fraction of the input units to drop in all the dropout layers in the network.
                Set this parameter to 0 to disable dropout (default).

        Returns:
            A ResNet1DArch object, which is a tensorflow model.
    """

    def __init__(self, n_classes, pre_trained_base=None, block_sets=None, initial_filters=16, 
                       initial_strides=1, initial_kernel=30, strides=2, kernel=300, 
                       batch_norm_momentum=0.99, dropout_rate=0, **kwargs):

        super(ResNet1DArch, self).__init__(**kwargs)

        self.n_classes = n_classes
        self.initial_strides = initial_strides
        self.initial_kernel = initial_kernel
        self.strides = strides
        self.kernel = kernel

        if pre_trained_base:
            self.conv_initial = pre_trained_base[0]
            self.blocks = pre_trained_base[1]
        
        else:
            self.n_sets = len(block_sets)
            self.block_sets = block_sets
            self.input_filters = initial_filters
            self.output_filters = initial_filters
            self.conv_initial = tf.keras.layers.Conv1D(filters=self.output_filters, kernel_size=initial_kernel, 
                                                strides=initial_strides, padding="same", use_bias=False,
                                                kernel_initializer=tf.random_normal_initializer())

            self.blocks = tf.keras.models.Sequential(name="dynamic_blocks")

            self.num_blocks = 0
            for set_id in range(self.n_sets):
                for block_id in range(self.block_sets[set_id]):
                    #First layer of every block except the first
                    if set_id != 0 and block_id == 0:
                        block = ResNet1DBlock(self.output_filters, strides=self.strides, kernel=self.kernel, 
                                residual_path=True, batch_norm_momentum=batch_norm_momentum, dropout_rate=dropout_rate)
                    
                    else:
                        if self.input_filters != self.output_filters:
                            residual_path = True
                        else:
                            residual_path = False

                        block = ResNet1DBlock(self.output_filters, strides=1, kernel=self.kernel,
                                residual_path=residual_path, batch_norm_momentum=batch_norm_momentum, dropout_rate=dropout_rate)

                    self.input_filters = self.output_filters

                    self.blocks.add(block)
                    self.num_blocks += 1
                
                self.output_filters *= 2

        self.batch_norm_final = tf.keras.layers.BatchNormalization(momentum=batch_norm_momentum)
        self.average_pool = tf.keras.layers.GlobalAveragePooling1D()
        self.fully_connected = tf.keras.layers.Dense(self.n_classes)
        self.softmax = tf.keras.layers.Softmax()
    
    
    def freeze_init_layer(self):
        """Freeze the initial convolutional layer"""
        self.layers[0].trainable = False

    def unfreeze_init_layer(self):
        """Unffreeze the initial convolutional layer"""
        self.layers[0].trainable = True
    
    def freeze_block(self, block_ids):
        """ Freeze specific residual blocks

            Args:
                blocks_ids: list of ints
                    The block numbers to be freezed (starting from zero)
        """

        for block_id in block_ids:
            self.layers[1].layers[block_id].trainable = False

    def unfreeze_block(self, block_ids):
        """ Unfreeze specific residual blocks

            Args:
                blocks_ids: list of ints
                    The block numbers to be freezed (starting from zero)
        """

        for block_id in block_ids:
            self.layers[1].layers[block_id].trainable = True
    
    def freeze_top(self):
        """Freeze the classification block"""
        for layer in self.layers[2:]:
            layer.trainable = False
    
    def unfreeze_top(self):
        """Unfreeze the classification block"""
        for layer in self.layers[2:]:
            layer.trainable = True

    def get_feature_extraction_base(self):
        """ Retrive the feature extraction base (initial convolutional layer + residual blocks)
        
            Returns:
                list containing the feature extraction layers
        """
        return [self.conv_initial, self.blocks]

    def clone_with_new_top(self, n_classes=None, freeze_base=True):
        """ Clone this instance but replace the original classification top with a new (untrained) one
        
            Args:
                n_classes:int
                    The number of classes the new classification top should output.
                    If None(default), the original number of classes will be used.
                freeze_base:bool
                    If True, the weights of the feature extraction base will be froze (untrainable) in the new model.
                
            Returns:
                cloned_model: instance of ResNetArch
                    The new model with the old feature extraction base and new classification top.
         """
        if freeze_base == True:
            self.trainable = False

        if n_classes is None:
            n_classes = self.n_classes

        pre_trained_base = self.get_feature_extraction_base()
        cloned_model = type(self)(n_classes=n_classes, pre_trained_base=pre_trained_base)

        return cloned_model

    def set_batch_norm_momentum(self, momentum):
        """ Set the momentum for the moving average of all the batch normalization layers in the network.

            For an explanation of how the momentum affects the batch normalisation operation,
            see <https://www.tensorflow.org/api_docs/python/tf/keras/layers/BatchNormalization>

            Args: 
                momentum: float between 0 and 1
                    Momentum for the moving average of the batch normalization layers.

            Returns:
                None
        """
        assert momentum>=0 and momentum<=1, 'batch normalization momentum must be between 0 and 1'

        self.batch_norm_final.momentum = momentum
        for block_id in range(self.num_blocks):
            self.layers[1].layers[block_id].set_batch_norm_momentum(momentum=momentum)

    def set_dropout_rate(self, rate):
        """ Set the fraction of the input units to drop in all the dropout layers in the network.

            Args: 
                rate: float between 0 and 1
                    Fraction of the input units to drop in the dropout layers.

            Returns:
                None
        """
        assert rate>=0 and rate<=1, 'dropout rate must be between 0 and 1'

        for block_id in range(self.num_blocks):
            self.layers[1].layers[block_id].set_dropout_rate(rate=rate)

    def call(self, inputs, training=None):
        """Calls the model on new inputs.

        In this case call just reapplies all ops in the graph to the new inputs (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: Tensor or list of tensors
                A tensor or list of tensors
            
            training: Bool
                Boolean or boolean scalar tensor, indicating whether to run the Network in training mode or inference mode.

        Returns:
                A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        output = self.call_frontend(inputs)
        output = self.conv_initial(output)
        output = self.blocks(output, training=training)
        output = self.batch_norm_final(output, training=training)
        output = tf.nn.relu(output)
        output = self.average_pool(output)
        output = self.fully_connected(output)
        output = self.softmax(output)

        return output




class ResNetInterface(NNInterface):
    """ Creates a ResNet model with the standardized Ketos interface.

        Args:
             block_sets: list of ints
                A list specifying the block sets and how many blocks each  set contains.
                Example: [2,2,2] will create a ResNet with 3 block sets, each containing
                2 ResNetBlocks (i.e.: a total of 6 residual blocks)
            
            n_classes:int
                The number of classes. The output layer uses a Softmax activation and
                will contain this number of nodes, resulting in model outputs with this
                many values summing to 1.0.

            initial_filters:int
                The number of filters used in the first ResNetBlock. Subsequent blocks 
                will have two times more filters than their previous block.

            initial_strides: int
                Strides used in the first convolutional layer

            initial_kernel: int
                Kernel size used in the first convolutional layer

            strides: int
                Strides used in convolutional layers within the blocks

            kernel: int
                Kernel size used in convolutional layers within the blocks

            optimizer: ketos.neural_networks.RecipeCompat object
                A recipe compatible optimizer (i.e.: wrapped by the ketos.neural_networksRecipeCompat class)

            loss_function: ketos.neural_networks.RecipeCompat object
                A recipe compatible loss_function (i.e.: wrapped by the ketos.neural_networksRecipeCompat class)

            metrics: list of ketos.neural_networks.RecipeCompat objects
                A list of recipe compatible metrics (i.e.: wrapped by the ketos.neural_networksRecipeCompat class).
                These metrics will be computed on each batch during training.

            secondary_metrics: list of ketos.neural_networks.RecipeCompat objects
                A list of recipe compatible metrics (i.e.: wrapped by the ketos.neural_networksRecipeCompat class).
                These can be used as additional metrics. Computed at each batch during training but only printed or
                logged as the average at the end of the epoch
                
    """
    def __init__(self, block_sets=default_resnet_recipe['block_sets'], 
                        n_classes=default_resnet_recipe['n_classes'], 
                        initial_filters=default_resnet_recipe['initial_filters'],
                        initial_strides=default_resnet_recipe['initial_strides'],
                        initial_kernel=default_resnet_recipe['initial_kernel'],
                        strides=default_resnet_recipe['strides'],
                        kernel=default_resnet_recipe['kernel'],
                        optimizer=default_resnet_recipe['optimizer'], 
                        loss_function=default_resnet_recipe['loss_function'], 
                        metrics=default_resnet_recipe['metrics']):
        super(ResNetInterface, self).__init__(optimizer, loss_function, metrics)
        self.block_sets = block_sets
        self.n_classes = n_classes
        self.initial_filters = initial_filters
        self.initial_strides = initial_strides
        self.initial_kernel  = initial_kernel
        self.strides = strides
        self.kernel  = kernel

        self.model=ResNetArch(block_sets=block_sets, n_classes=n_classes, initial_filters=initial_filters,
                                initial_strides=initial_strides, initial_kernel=initial_kernel, 
                                strides=strides, kernel=kernel)

    @classmethod
    def _build_from_recipe(cls, recipe, recipe_compat=True):
        """ Build a ResNet model from a recipe.

            Args:
                recipe: dict
                    A recipe dictionary. optimizer, loss function
                    and metrics must be instances of ketos.neural_networks.RecipeCompat.
                    
                    Example recipe (minimal):
                    
                    >>> {{'block_sets':[2,2,2], # doctest: +SKIP
                    ...    'n_classes':2,
                    ...    'initial_filters':16,        
                    ...    'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                    ...    'loss_function': RecipeCompat('FScoreLoss', FScoreLoss),  
                    ...    'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)],
                    }

                    Example recipe (full):
                    
                    >>> {{'block_sets':[2,2,2], # doctest: +SKIP
                    ...    'n_classes':2,
                    ...    'initial_filters':16,   
                    ...     initial_strides':1,     
                    ...     initial_kernel':[3,3],     
                    ...     strides':2,     
                    ...     kernel':[3,3],     
                    ...    'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                    ...    'loss_function': RecipeCompat('FScoreLoss', FScoreLoss),  
                    ...    'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)],
                    }                     

            Returns:
                An instance of ResNetInterface.

        """
        block_sets = recipe['block_sets']
        n_classes = recipe['n_classes']
        initial_filters = recipe['initial_filters']

        initial_strides = recipe['initial_strides'] if 'initial_strides' in recipe.keys() else default_resnet_recipe['initial_strides']
        initial_kernel = recipe['initial_kernel'] if 'initial_kernel' in recipe.keys() else default_resnet_recipe['initial_kernel']
        strides = recipe['strides'] if 'strides' in recipe.keys() else default_resnet_recipe['strides']
        kernel = recipe['kernel'] if 'kernel' in recipe.keys() else default_resnet_recipe['kernel']
        
        if recipe_compat == True:
            optimizer = recipe['optimizer']
            loss_function = recipe['loss_function']
            metrics = recipe['metrics']
            
        else:
            optimizer = cls._optimizer_from_recipe(recipe['optimizer'])
            loss_function = cls._loss_function_from_recipe(recipe['loss_function'])
            metrics = cls._metrics_from_recipe(recipe['metrics'])
            
        instance = cls(block_sets=block_sets, n_classes=n_classes, initial_filters=initial_filters,
                        initial_strides=initial_strides, initial_kernel=initial_kernel, strides=strides, kernel=kernel, 
                        optimizer=optimizer, loss_function=loss_function, metrics=metrics)

        return instance

    @classmethod
    def _read_recipe_file(cls, json_file, return_recipe_compat=True):
        """ Read a ResNet recipe saved in a .json file.

            Args:
                json_file:string
                    Full path (including filename and extension) to the .json file containing the recipe.
                return_recipe_compat:bool
                    If True, returns a dictionary where the optimizer, loss_function, metrics and 
                    secondary_metrics (if available) values are instances of the ketos.neural_networks.nn_interface.RecipeCompat.
                    The returned dictionary will be equivalent to:
                            
                            >>> {'block_sets':[2,2,2], # doctest: +SKIP
                            ... 'n_classes':2,
                            ... 'initial_filters':16,        
                            ... 'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                            ... 'loss_function': RecipeCompat('FScoreLoss', FScoreLoss),  
                            ... 'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)]}

                    If False, the optimizer, loss_function, metrics and secondary_metrics (if available) values will contain a
                    dictionary representation of such fields instead of the RecipeCompat objects:
                            >>> {'block_sets':[2,2,2], # doctest: +SKIP
                            ... 'n_classes':2,
                            ... 'initial_filters':16,        
                            ... 'optimizer': {'name':'Adam', 'parameters': {'learning_rate':0.005}},
                            ... 'loss_function': {'name':'FScoreLoss', 'parameters':{}},  
                            ... 'metrics': [{'name':'CategoricalAccuracy', 'parameters':{}}]}

                Returns:
                    recipe, according to 'return_recipe_compat.

        """

        with open(json_file, 'r') as json_recipe:
            recipe_dict = json.load(json_recipe)

        optimizer = cls._optimizer_from_recipe(recipe_dict['optimizer'])
        loss_function = cls._loss_function_from_recipe(recipe_dict['loss_function'])
        metrics = cls._metrics_from_recipe(recipe_dict['metrics'])
        
        if return_recipe_compat == True:
            recipe_dict['optimizer'] = optimizer
            recipe_dict['loss_function'] = loss_function
            recipe_dict['metrics'] = metrics
            
        else:
            recipe_dict['optimizer'] = cls._optimizer_to_recipe(optimizer)
            recipe_dict['loss_function'] = cls._loss_function_to_recipe(loss_function)
            recipe_dict['metrics'] = cls._metrics_to_recipe(metrics)
        
        recipe_dict['block_sets'] = recipe_dict['block_sets']
        recipe_dict['n_classes'] = recipe_dict['n_classes']
        recipe_dict['initial_filters'] = recipe_dict['initial_filters']

        recipe_dict['initial_strides'] = recipe_dict['initial_strides'] if 'initial_strides' in recipe_dict.keys() else default_resnet_recipe['initial_strides']
        recipe_dict['initial_kernel'] = recipe_dict['initial_kernel'] if 'initial_kernel' in recipe_dict.keys() else default_resnet_recipe['initial_kernel']
        recipe_dict['strides'] = recipe_dict['strides'] if 'strides' in recipe_dict.keys() else default_resnet_recipe['strides']
        recipe_dict['kernel'] = recipe_dict['kernel'] if 'kernel' in recipe_dict.keys() else default_resnet_recipe['kernel']

        return recipe_dict

    def _extract_recipe_dict(self):
        """ Create a recipe dictionary from a ResNetInterface instance.

            The resulting recipe contains all the fields necessary to build the same network architecture used by the instance calling this method.
            
            Returns:
                recipe:dict
                    A dictionary containing the recipe fields necessary to build the same network architecture.
                    The output is equivalent to:
                        >>> {'block_sets':[2,2,2], # doctest: +SKIP
                        ...    'n_classes':2,
                        ...    'initial_filters':16,        
                        ...    'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                        ...    'loss_function': RecipeCompat('FScoreLoss', FScoreLoss),  
                        ...    'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)]}
        """

        recipe = {}
        recipe['interface'] = type(self).__name__
        recipe['block_sets'] = self.block_sets
        recipe['n_classes'] = self.n_classes
        recipe['initial_filters'] = self.initial_filters
        recipe['initial_strides'] = self.initial_strides
        recipe['initial_kernel'] = self.initial_kernel
        recipe['strides'] = self.strides
        recipe['kernel'] = self.kernel
        recipe['optimizer'] = self._optimizer_to_recipe(self.optimizer)
        recipe['loss_function'] = self._loss_function_to_recipe(self.loss_function)
        recipe['metrics'] = self._metrics_to_recipe(self.metrics)
        
        return recipe



class ResNet1DInterface(ResNetInterface): 
    @classmethod
    def transform_batch(cls, x, y, n_classes=2):
        """ Transforms a training batch into the format expected by the network.

            When this interface is subclassed to make new neural_network classes, this method can be overwritten to
            accomodate any transformations required. Common operations are reshaping of input arrays and parsing or one hot encoding of the labels.

            Args:
                x:numpy.array
                    The batch of inputs with shape (batch_size, width, height)
                y:numpy.array
                    The batch of labels.
                    Each label must be represented as an integer, ranging from zero to n_classes
                    The array is expected to have a field named 'label'.
                n_classes:int
                    The number of possible classes for one hot encoding.

            Returns:
                X:numpy.array
                    The transformed batch of inputs
                Y:numpy.array
                    The transformed batch of labels

            Examples:
                >>> import numpy as np
                >>> # Create a batch of 10 5x5 arrays
                >>> inputs = np.random.rand(10,5,5)
                >>> inputs.shape
                (10, 5, 5)

                    
                >>> # Create a batch of 10 labels (0 or 1)
                >>> labels = np.random.choice([0,1], size=10)
                >>> labels.shape
                (10,)

                >>> transformed_inputs, transformed_labels = NNInterface.transform_batch(inputs, labels, n_classes=2)
                >>> transformed_inputs.shape
                (10, 5, 5, 1)

                >>> transformed_labels.shape
                (10, 2)
                
        """
        X = cls._transform_input(x)
        if y.dtype.names is not None:
            Y = np.array([cls._to1hot(class_label=label, n_classes=n_classes) for label in y['label']])  
        else:    
            Y = np.array([cls._to1hot(class_label=label, n_classes=n_classes) for label in y])        

        return (X,Y)

    @classmethod
    def _transform_input(cls,input):
        """ Transforms a training input to the format expected by the network.

            Similar to :func:`NNInterface.transform_train_batch`, but only acts on the inputs (not labels). Mostly used for inference, rather than training.
            When this interface is subclassed to make new neural_network classes, this method can be overwritten to
            accomodate any transformations required. Common operations are reshaping of an input.

            Args:
                input:numpy.array
                    An input instance. Must be of shape (n,m) or (k,n,m).

            Raises:
                ValueError if input does not have 2 or 3 dimensions.

            Returns:
                tranformed_input:numpy.array
                    The transformed batch of inputs

            Examples:
                >>> import numpy as np
                >>> # Create a batch of 10 5x5 arrays
                >>> batch_of_inputs = np.random.rand(10,5,5)
                >>> selected_input = batch_of_inputs[0]
                >>> selected_input.shape
                (5, 5)
                 
                >>> transformed_input = NNInterface._transform_input(selected_input)
                >>> transformed_input.shape
                (1, 5, 5, 1)

                # The input can also have shape=(1,n,m)
                >>> selected_input = batch_of_inputs[0:1]
                >>> selected_input.shape
                (1, 5, 5)
                 
                >>> transformed_input = NNInterface._transform_input(selected_input)
                >>> transformed_input.shape
                (1, 5, 5, 1)

                
        """
        if input.ndim == 1:
            transformed_input = input.reshape(1,input.shape[0],1)
        elif input.ndim == 2:
            transformed_input = input.reshape(input.shape[0],input.shape[1],1)
        else:
            raise ValueError("Expected input to have 1 or 2 dimensions, got {}({}) instead".format(input.ndims, input.shape))

        return transformed_input

    def __init__(self, block_sets=default_resnet_1d_recipe['block_sets'], n_classes=default_resnet_1d_recipe['n_classes'], 
                        initial_filters=default_resnet_1d_recipe['initial_filters'],
                        initial_strides=default_resnet_1d_recipe['initial_strides'], initial_kernel=default_resnet_1d_recipe['initial_kernel'],
                        strides=default_resnet_1d_recipe['strides'], kernel=default_resnet_1d_recipe['kernel'], 
                        optimizer=default_resnet_1d_recipe['optimizer'], loss_function=default_resnet_1d_recipe['loss_function'], 
                        metrics=default_resnet_1d_recipe['metrics']):

        super(ResNet1DInterface, self).__init__(optimizer=optimizer, loss_function=loss_function, metrics=metrics)

        self.block_sets = block_sets
        self.n_classes = n_classes
        self.initial_filters = initial_filters
        self.initial_strides = initial_strides
        self.initial_kernel  = initial_kernel
        self.strides = strides
        self.kernel  = kernel
       
        self.model=ResNet1DArch(block_sets=block_sets, n_classes=n_classes, initial_filters=initial_filters, 
                                initial_strides=initial_strides, initial_kernel=initial_kernel, strides=strides, kernel=kernel)
       



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" detection sub-module within the ketos.neural_networks.dev_utils module

    This module provides auxiliary functions to incorporate pre-trained ketos classifiers models into detection tools

    Contents:
        

"""

import numpy as np
import pandas as pd
import os
import math
import warnings
import logging
import tables
from operator import itemgetter
from itertools import groupby
from tqdm import tqdm, trange
from scipy.ndimage import convolve

def compute_score_running_avg(scores, window_size):
    """
    This function calculates the running average of a given list of scores, over a defined window size.

    Each element i in the output represents the running average of the elements 
    in the scores from position i - window_size//2 to i + window_size//2.

    This function pads the scores with the edge values of input to calculate the average for the edges.
    
    Args:
        scores: list or numpy array
            A 1D, 2D or 3D sequence of numerical scores.
        window_size: int 
            The size of the window in frames to compute the running average. Must be an odd integer

    Returns:
        numpy array
            A sequence of running averages, with the same length as the input. 
    
    Example:

    >>> compute_score_running_avg([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 3)
    array([1.33333333, 2.        , 3.        , 4.        , 5.        ,
           6.        , 7.        , 8.        , 9.        , 9.66666667])

    >>> compute_score_running_avg([[1, 1], [2, 2], [3, 3], [4, 4], [5, 5]], 3)
    array([[1.33333333, 1.33333333],
           [2.        , 2.        ],
           [3.        , 3.        ],
           [4.        , 4.        ],
           [4.66666667, 4.66666667]])
    
    >>> np.random.seed(0)
    >>> scores = np.random.rand(6, 4)
    >>> compute_score_running_avg(scores, 3)
    array([[0.50709394, 0.69209095, 0.54770465, 0.66051312],
           [0.64537702, 0.58150833, 0.61069188, 0.6551837 ],
           [0.65178737, 0.65164409, 0.43344944, 0.50259907],
           [0.51730857, 0.713886  , 0.54697262, 0.49534546],
           [0.52229377, 0.85245835, 0.43689072, 0.57922354],
           [0.65915169, 0.81031232, 0.56703849, 0.81035683]])
    """

    if window_size % 2 == 0:
        raise ValueError("window_size must be an odd integer")
    
    # Ensure scores is a numpy array for convenience
    scores = np.array(scores, dtype=float)
    
    # Check if window_size is larger than the length of input scores
    if window_size > scores.shape[0]:
        raise ValueError("window_size cannot be larger than the length of the input scores")

    
    # Ensure we have 3 or less dimensions
    if scores.ndim > 4:
        raise ValueError("scores must be a 1D, 2D or 3D list or array")

    weights = np.ones(window_size) / window_size
    # here we are creating the weights which can be viwed as doing the divison operation of calculating the avarage before actually summing the values
    # The weights will be applied to each input in the window and summed together in the convolution operation
    if scores.ndim == 1:
        return convolve(scores, weights, mode='nearest')
    elif scores.ndim == 2 or scores.ndim == 3:
        return np.apply_along_axis(convolve, axis=0, arr=scores, weights=weights, mode='nearest')
    else:
        raise ValueError("scores must be a 1D, 2D, or 3D list or array")

def batch_load_hdf5_data(hdf5_file_path, batch_size, audio_representation, start_idx=0, table_name=None, x_field='/data', logger=None):
    """
    Loads HDF5 data in batches from a specified file and yields it for processing.

    This function assumes that the HDF5 dataset is named 'data'. For each batch, it will yield a dictionary
    containing the batch's data, filename, start and end offsets.

    - The 'data' key in the yielded dictionary corresponds to the batch's data.
    - The 'filename' key in the yielded dictionary corresponds to the batch's filename.
    - The 'start' key in the yielded dictionary corresponds to the start offset of the batch.
    - The 'end' key in the yielded dictionary corresponds to the end offset of the batch.

    Args:
        hdf5_file_path : str
            The file path of the HDF5 file to read from.
        batch_size : int
            The number of samples to include in each batch.
        audio_representation : dict
            A dictionary containing information about the audio representation.
        start_idx: int
            The batch index to start from. This allows the generator to skip the initial batches up to the specified index.
        table_name : str
            The name of the table in the HDF5 file to read the data from.
            If no name is provided, it defaults to '/data'.
        x_field : str
            Name of the field in the table to access the data.
        logger: logging.Logger or KetosLogger
            A Logger instance to log errors encountered while loading audio file data.

    Returns:
        dict
            A dictionary containing the batch data, its filename, start and end offsets.
            The 'end' key value is calculated based on the 'duration' key of the audio_representation parameter
            and may not be accurate, as the duration might change when actually computing the representation.
    """
    if table_name is None:
        table_name = "/data"
    else:
        table_name = table_name + "/data"

    # Define a vectorized function to decode bytes to utf-8 strings
    decode_utf8 = np.vectorize(lambda x: x.decode('utf-8'))

    with tables.open_file(hdf5_file_path, 'r') as hdf5_file:
        # Assuming the dataset is named 'data'
        table = hdf5_file.get_node(table_name)
        num_samples = table.shape[0]
        num_batches = int(np.ceil(num_samples / batch_size))

        batch_data = {}
        for i in trange(start_idx, num_batches):
            try:
                start = i * batch_size
                # Here we are getting the num of samples in a batch. Usually it is the batch size except for the last batch which could be smaller
                end = min((i + 1) * batch_size, num_samples)
                hdf5_batch_data = table[start:end]

                batch_data['data'] = hdf5_batch_data[x_field]
                batch_data['filename'] = decode_utf8(hdf5_batch_data['filename'])
                batch_data['start'] = hdf5_batch_data['offset']
                batch_data['end'] = hdf5_batch_data['offset'] + audio_representation['duration'] # this isnt quite right because the duration may change when actually computing the representation
            except Exception as e:
                print(e)
                if logger is not None:
                    logger.error(f'{e} Skipping File.', stdout=True, **{'status': 'error'})
                continue  # Skip this iteration if an error occurred

            yield batch_data

def batch_load_audio_file_data(loader, batch_size, start_idx=0, logger=None):
    """
    This function generates batches of audio data from a given AudioFrameLoader.

    Each batch consists of spectrogram data, filename, start time and end time of each audio segment.
    The loader loads audio files, splits into smaller segments and convert them into spectrograms.

    Args:
        loader: ketos.audio.audio_loader.AudioFrameLoader
            An AudioFrameLoader object that computes spectrograms from the audio files as requested.
        batch_size: int 
            The number of samples to include in each batch.
        start_idx: int
            The batch index to start from. This allows the generator to skip the initial batches up to the specified index.
        logger: logging.Logger or KetosLogger
            A Logger instance to log errors encountered while loading audio file data.
            

    Returns:
        dict: A dictionary containing the batch data. The dictionary keys are 'data', 'filename', 
              'start', and 'end'. The 'data' field is a numpy array of shape (batch_size, time_bins, freq_bins), 
              where time_bins and freq_bins correspond to the shape of the spectrogram data.
              'filename' contains the list of filenames for the audio segments in the batch.
              'start' and 'end' are lists containing the start and end times (in seconds) of each audio segment.
    """
    n_batches = int(np.ceil(loader.num() / batch_size))

    # Skip the previously processed batches
    for _ in range(start_idx * batch_size):
        loader.skip()

    for batch_id in trange(start_idx, n_batches):
        # This will compute how many samples should be loaded in the batch. Normally it is the batch size, 
        # but in the case where the last batch is smaller than batch size we just load the necessary amount
        # Each sample is a segment created by the AudioFrameLoader
        num_samples = min(batch_size, loader.num() - batch_size * batch_id) 

        batch_data = {"data": [], "filename": [], "start": [], "end": []}
        for _ in range(num_samples):
            try:
                spec = next(loader)

                # for some reason, the actual spectrogram data is found in data.data We should change this at some point
                # Batch data will be a list with length batch_size with each item beeing a np.array with shape (time_bins, freq_bins)
                batch_data['data'].append(spec.get_data()) 
                batch_data['filename'].append(spec.filename)
                batch_data['start'].append(spec.start)
                batch_data['end'].append(spec.start + float(spec.duration()))
            except Exception as e:
                if logger is not None:
                    logger.error(f'{e} Skipping File.', stdout=True, **{'status': 'error'})
                continue  # Skip this iteration if an error occurred
        # transform batch_data to be a numpy array with shape (batch_size, time_bins, freq_bins)
        batch_data['data'] = np.stack(batch_data['data'], axis=0) 
        yield batch_data

def add_detection_buffer(detections_df, buffer):
    """ Add a buffer to each detection in the DataFrame.

        Args:
            detections_df: pandas DataFrame
                DataFrame with detections. It should have the following columns:
                - 'filename': The name of the file containing the detection.
                - 'start': The start time of the detection in seconds.
                - 'end': The end time of the detection in seconds.
                - 'score': The score associated with the detection.
            buffer: float
                The buffer duration to be added to each detection in seconds.
        
        Returns:
            detections_df: pandas DataFrame
                DataFrame with the detections after adding the buffer.

        Example:
            Given a step_size of 0.5 and a DataFrame with the following format:

            +----------+-------+-----+-------+
            | filename | start | end | score |
            +----------+-------+-----+-------+
            | file1    | 1     | 3   | 0.9   |
            +----------+-------+-----+-------+
            | file2    | 0     | 2   | 0.8   |
            +----------+-------+-----+-------+

            The function would return:

            +----------+-------+-----+-------+
            | filename | start | end | score |
            +----------+-------+-----+-------+
            | file1    | 0.5   | 3.5 | 1.5   |
            +----------+-------+-----+-------+
            | file2    | 0     | 2.5 | 3     |
            +----------+-------+-----+-------+

        >>> import pandas as pd
        >>> detections_df = pd.DataFrame([
        ...     {'filename': 'file1', 'start': 1, 'end': 3, 'score': 0.9},
        ...     {'filename': 'file2', 'start': 0, 'end': 2, 'score': 0.8},
        ... ])
        >>> buffer = 0.5
        >>> add_detection_buffer(detections_df, buffer)
          filename  start  end  score
        0    file1    0.5  3.5    0.9
        1    file2    0.0  2.5    0.8

        >>> detections_df = pd.DataFrame([
        ...     {'filename': 'file1', 'start': 0, 'end': 5, 'score': 1.0},
        ...     {'filename': 'file2', 'start': 2, 'end': 4, 'score': 0.7},
        ... ])
        >>> buffer = 1.0
        >>> add_detection_buffer(detections_df, buffer)
          filename  start  end  score
        0    file1    0.0  6.0    1.0
        1    file2    1.0  5.0    0.7
    """
    detections_df_copy = detections_df.copy()

    detections_df_copy['start'] = detections_df_copy['start'].apply(lambda x: max(0, x - buffer)) # detection start cant be lower than 0
    detections_df_copy['end'] += buffer

    return detections_df_copy


def apply_detection_threshold(scores, threshold=0.5, highest_score_only=False):
    """
    Filters out detection scores below or at a specified threshold and returns a list of tuples, where each tuple consists of a label (the index of the score) and the score itself.
    
    Args:
        scores: list of floats
            The list of scores.
        threshold: float
            The threshold below which scores are filtered out. Default is 0.5.
        highest_score_only: bool
            If True, only the highest score is returned. Default is False.

    Returns:
        list of tuples: Each tuple contains a label (the index of the score in the input list) and the score itself.

    Examples:

    >>> apply_detection_threshold([0.2, 0.7, 0.3], 0.5)
    [(1, 0.7)]
    >>> apply_detection_threshold([0.6, 0.4, 0.8], 0.55)
    [(0, 0.6), (2, 0.8)]
    >>> apply_detection_threshold([0.6, 0.4, 0.8], 0.6, True)
    [(2, 0.8)]
    """

    filtered_scores = [(label, score) for label, score in enumerate(scores) if score >= threshold]
    
    if highest_score_only and filtered_scores:
        return [max(filtered_scores, key=lambda item: item[1])]
    else:
        return filtered_scores

def filter_by_threshold(detections, threshold=0.5, highest_score_only=False):
    """
    Filters out detection scores below a specified threshold and returns a DataFrame with the remaining detections. 
    
    Args:
        detections: dict
            The dictionary with the detections.
        threshold: float
            The threshold below which scores are filtered out. Default is 0.5.
        highest_score_only: bool
            If True, only the highest score is returned. Default is False.

    Returns:
        pd.DataFrame: DataFrame with the remaining detections.

    Examples:
    
    >>> detections = {
    ...    'filename': ['file1', 'file2'],
    ...    'start': [0, 1],
    ...    'end': [1, 2],
    ...    'score': [[0.2, 0.7, 0.3], [0.1, 0.4, 0.8]]
    ... }
    >>> filter_by_threshold(detections, 0.5)
      filename  start  end  label  score
    0    file1      0    1      1    0.7
    1    file2      1    2      2    0.8
    """
    # create an empty list to store the filtered output
    filtered_output = {'filename': [], 'start': [], 'end': [], 'label': [], 'score': []}

    for filename, start, end, scores in zip(detections['filename'], detections['start'], detections['end'], detections['score']):
        detections = apply_detection_threshold(scores, threshold, highest_score_only)
        # for each score that passes the threshold, duplicate the filename, start, end, and add the corresponding label
        # if there is no score above the threshold, exclude the segment from the 
        for label, score in detections:
            filtered_output['filename'].append(filename)
            filtered_output['start'].append(start)
            filtered_output['end'].append(end)
            filtered_output['label'].append(label)
            filtered_output['score'].append(float(score))
    df = pd.DataFrame(filtered_output)
    return df



def convert_sequence_to_snapshot(detections, threshold=0.5, highest_score_only=False):
    """
    Converts a sequence of scores into a snapshot of events that exceed a threshold.

    This assumes the output is from a sequence model. If using a snapshot model already, see :meth:`neural_networks.dev_utils.detection.filter_by_threshold`,
    
    The input data should be a dictionary containing:
      'filename': a list of filenames,
      'start': a list of start times for each input,
      'end': a list of end times for each input,
      'score': a 3D array or list of scores for each input, each sublist corresponding to a class.

    The output is a dictionary that contains:
      'filename': the filename where the event occurred,
      'start': the start time of the event,
      'end': the end time of the event,
      'label': the label of the class that triggered the event,
      'score': the scores of the class that triggered the event.

    Args:
        detections: dict
            A dictionary containing the filenames, start times, end times, and scores.
        threshold: float
            A threshold value to filter scores. Scores above this threshold are considered as an event. Default is 0.5.
        highest_score_only: bool
            If True, only the highest score is returned. Default is False.

    Returns:
        pandas.DataFrame: A DataFrame containing the snapshots of events.

    Examples:

    >>> data = {
    ...     'filename': ['file1.wav'],
    ...     'start': [0.0],
    ...     'end': [60.0],
    ...     'score': [[
    ...         [0.1, 0.6, 0.4],
    ...         [0.2, 0.3, 0.5]
    ...     ]]
    ... }
    >>> df = convert_sequence_to_snapshot(data, 0.5)
    >>> df.equals(pd.DataFrame({
    ...     'filename': ['file1.wav'], 
    ...     'start': [20.0], 
    ...     'end': [40.0], 
    ...     'label': [0], 
    ...     'score': [[0.6]]
    ... }))
    True
    >>> data = {
    ...     'filename': ['file1.wav'],
    ...     'start': [0.0],
    ...     'end': [60.0],
    ...     'score': [[
    ...         [0.8, 0.9, 0.7, 0.2, 0.1, 0.6, 0.9, 0.9, 0.2],
    ...         [0.7, 0.4, 0.6, 0.8, 0.6, 0.7, 0.8, 0.6, 0.1]
    ...     ]]
    ... }
    >>> df = convert_sequence_to_snapshot(data, 0.5, highest_score_only=True)
    >>> df = df.round(5) 
    >>> df.equals(pd.DataFrame({
    ...     'filename': ['file1.wav', 'file1.wav', 'file1.wav'], 
    ...     'start': [0.0, 20.0, 40.0], 
    ...     'end': [20.0, 40.0, 53.33333], 
    ...     'label': [0, 1, 0], 
    ...     'score': [[0.8, 0.9, 0.7], [0.8, 0.6, 0.7], [0.9, 0.9]]
    ... }))
    True
    """
    filtered_output = {'filename': [], 'start': [], 'end': [], 'label': [], 'score': []}
    
    for i in range(len(detections['filename'])):
        filename = detections['filename'][i]
        start = detections['start'][i]
        end = detections['end'][i]
        scores_list = detections['score'][i]
        
        # convert the score list into a numpy array for efficient computation
        scores = np.array(scores_list)
        # create mask where scores is above the threshold
        mask = scores > threshold

        if highest_score_only:
            # create mask where we keep the maximum between the classes
            mask2 = scores.max(axis=0, keepdims=1) == np.array(scores_list)
            # the final mask will be scores above the threshold and only the maximum score for that particular segment out of the classes
            mask = np.logical_and(mask, mask2)

        for label, score_list in enumerate(scores):
            # apply the mask to the score list for this particular class and get the index where the value is true
            filtered_indexes = np.where(mask[label] == True)[0]

            if filtered_indexes.size > 0:
                # grouping consecutive numbers
                for k, g in groupby(enumerate(filtered_indexes), key=lambda x: x[0] - x[1]):
                    group = list(map(itemgetter(1), g)) # gets the original index (without the position from enumerate)
                    start_time = start + group[0] / len(score_list) * (end - start) # here we are converting frames back into seconds.
                    end_time = start + (group[-1] + 1) / len(score_list) * (end - start)  # plus one because end time should be exclusive
                    filtered_output['filename'].append(filename)
                    filtered_output['start'].append(start_time)
                    filtered_output['end'].append(end_time)
                    filtered_output['label'].append(label)
                    filtered_output['score'].append(score_list[group].tolist())
    
    df = pd.DataFrame(filtered_output)
    
    return df.sort_values(by=['start'], ignore_index=True)

def filter_by_label(detecitons, labels):
    """
    Filters the input DataFrame by specified label(s).
    
    Args:
        detections: pandas DataFrame
            A DataFrame containing the results data.
        labels: list or integer 
            A list of labels to filter by.

    Returns:
        pandas.DataFrame: 
            A DataFrame containing only the detections with the specified labels.

    Example:

    >>> import pandas as pd
    >>> df = pd.DataFrame({
    ...     'filename': ['file1.wav', 'file2.wav', 'file3.wav'],
    ...     'start': [0.0, 20.0, 40.0],
    ...     'end': [20.0, 40.0, 60.0],
    ...     'label': [0, 1, 2],
    ...     'score': [0.6, 0.8, 0.7]
    ... })
    >>> filtered_df = filter_by_label(df, 1)
    >>> filtered_df = filtered_df.reset_index(drop=True)
    >>> filtered_df.equals(pd.DataFrame({
    ...     'filename': ['file2.wav'],
    ...     'start': [20.0],
    ...     'end': [40.0],
    ...     'label': [1],
    ...     'score': [0.8]
    ... }))
    True

    """
    if not isinstance(labels, list):
        labels = [labels]
    return detecitons[detecitons['label'].isin(labels)]

def merge_overlapping_detections(detections_df):
    """ Merge overlapping or adjacent detections with the same label.

        The score of the merged detection is computed as the average of the individual detection scores.

        Note: The detections are assumed to be sorted by start time in chronological order.

        Args:
            detections_df: pandas DataFrame
                Dataframe with detections. It should have the following columns:
                - 'filename': The name of the file containing the detection.
                - 'start': The start time of the detection in seconds.
                - 'end': The end time of the detection in seconds.
                - 'label': The label associated with the detection.
                - 'score': The score associated with the detection.
        
        Returns:
            merged: pandas DataFrame
                DataFrame with the merged detections.

        Example:
            Given a DataFrame with the following format:

            +----------+-------+-----+-------+-------+
            | filename | start | end | label | score |
            +----------+-------+-----+-------+-------+
            | file1    | 0     | 5   | 0     | 1     |
            +----------+-------+-----+-------+-------+
            | file1    | 3     | 7   | 0     | 2     |
            +----------+-------+-----+-------+-------+
            | file2    | 0     | 5   | 1     | 3     |
            +----------+-------+-----+-------+-------+

            The function would return:
            
            +----------+-------+-----+-------+-------+
            | filename | start | end | label | score |
            +----------+-------+-----+-------+-------+
            | file1    | 0     | 7   | 0     | 1.5   |
            +----------+-------+-----+-------+-------+
            | file2    | 0     | 5   | 1     | 3     |
            +----------+-------+-----+-------+-------+

        >>> import pandas as pd
        >>> detections_df = pd.DataFrame([
        ...     {'filename': 'file1', 'start': 0, 'end': 5, 'label': 0, 'score': 1},
        ...     {'filename': 'file1', 'start': 3, 'end': 7, 'label': 0, 'score': 2},
        ...     {'filename': 'file2', 'start': 0, 'end': 5, 'label': 1, 'score': 3}
        ... ])
        >>> merged = merge_overlapping_detections(detections_df)
        >>> merged.to_dict('records')
        [{'filename': 'file1', 'start': 0, 'end': 7, 'label': 0, 'score': 1.5}, {'filename': 'file2', 'start': 0, 'end': 5, 'label': 1, 'score': 3.0}]
    """
    detections = detections_df.to_dict('records')

    if len(detections) <= 1:
        return detections_df
    
    merged_detections = [detections[0]]

    for i in range(1,len(detections)):
        # detections do not overlap, nor are they adjacent nor they are from the same label
        if detections[i]['start'] > merged_detections[-1]['end'] or detections[i]['filename'] != merged_detections[-1]['filename'] or detections[i]['label'] != merged_detections[-1]['label']:
            merged_detections.append(detections[i])
        # detections overlap, or adjacent to one another
        else:
            # determine if the score is a list or a single number
            if isinstance(merged_detections[-1]['score'], list):
                overlap = merged_detections[-1]['end'] - detections[i]['start'] + 1 # amount of overlap

                # handle the scores
                merged_scores = merged_detections[-1]['score'][: -overlap]  # get the non-overlapping part from the first detection
                overlap_scores_merged = merged_detections[-1]['score'][-overlap:]  # get the overlapping part from the first detection
                overlap_scores_new = detections[i]['score'][:overlap]  # get the overlapping part from the second detection
                overlap_scores_avg = [(x + y) / 2 for x, y in zip(overlap_scores_merged, overlap_scores_new)]  # average the overlapping scores
                
                merged_scores += overlap_scores_avg  # add the averaged overlapping scores to the merged scores
                merged_scores += detections[i]['score'][overlap:]  # add the non-overlapping part from the second detection to the merged scores
            else:
                # if score is a single number, just average it as before
                merged_scores = (merged_detections[-1]['score'] + detections[i]['score']) / 2
            
            # create the new merged detection
            merged_detection = {
                'filename': detections[i]['filename'], 
                'start': merged_detections[-1]['start'], 
                'end': detections[i]['end'], 
                'label': detections[i]['label'], # add the label of the merged detection
                'score': merged_scores
            }

            merged_detections[-1] = merged_detection #replace

    return pd.DataFrame(merged_detections)
    
def merge_consecutive_detections(detections_df, step_size):
    # This implementation has several questions:

    # Consider if we really need this

    # 1) what happens when there are for instance, two consecutive detections for instance, from 0-3 and 1-4 and then one from 3-6? Does this create 2 detections or 1 detection? The current implementation creates 2
    # 2) See this: https://git-dev.cs.dal.ca/meridian/ketos/-/issues/18
    # 4) what if the step size is greater than the duration?
    """ Merges consecutive detections in the given dataframe.

        Consecutive detections are merged into a single detection event represented by the time interval start-end. 
        Consecutive detections are determined by the step_size parameter. The start time of the merged detection 
        will be the start time of the first consecutive detection. The end time of the merged detection will be
        adjusted to be 'step_size units after the start of the last consecutive detection.
        
        If the detections are not consecutive, each detection will have its end time adjusted to be 'step_size units after the start.
    
        See examples below.

        Args:
            detections_df: pandas DataFrame
                Dataframe with detections. It should have the following columns:
                - 'filename': The name of the file containing the detection.
                - 'start': The start time of the detection in seconds.
                - 'end': The end time of the detection in seconds.
                - 'label': The label associated with the detection.
                - 'score': The score associated with the detection.
            step_size: float
                The time interval (in seconds) between the starts of each continuous inputs.
                For example, a step=0.5 indicates that the first spectrogram starts at time 0.0s (from the beginning of the audio file), the second at 0.5s, etc.

        Returns:
            merged: pandas DataFrame
                DataFrame with the merged detections.

        Examples:

        >>> import pandas as pd
        >>> detections_df = pd.DataFrame({
        ...     'filename': ['file1', 'file1', 'file2', 'file2'],
        ...     'start': [0.0, 2.0, 0.0, 10.0],
        ...     'end': [3.0, 5.0, 3.0, 13.0],
        ...     'label': [1, 1, 1, 1],
        ...     'score': [0.8, 0.6, 0.9, 0.7]
        ... })
        >>> step_size = 2.0
        >>> merged = merge_consecutive_detections(detections_df, step_size)
        >>> merged
          filename  start   end  label  score
        0    file1    0.0   4.0      1    0.7
        1    file2    0.0   2.0      1    0.9
        2    file2   10.0  12.0      1    0.7
    """
    if len(detections_df) == 0:
        return detections_df
    
    if step_size < 0:
        raise ValueError("step_size must be non-negative")
    
    merged_detections = []
    for _, group in detections_df.groupby('filename'):
        group = group.to_dict('records') # transforming to dict for faster and easier looping

        merged_detection = {'filename': group[0]['filename'], 'start': group[0]['start'], 'end': group[0]['start'] + step_size, 'label': group[0]['label'], 'score': group[0]['score']}
        consecutives = 1 # start a counter of consecutive detections

        for i in range(1, len(group)):
            # detections are consecutive and have the same label
            if math.isclose(group[i-1]['start'] + step_size, group[i]['start']) and group[i-1]['label'] == group[i]['label']:
                consecutives += 1

                if isinstance(group[-1]['score'], list):
                    merged_detection['score'] = [sum(x) for x in zip(merged_detection['score'], group[i]['score'])]
                else:
                    merged_detection['score'] += group[i]['score']

                merged_detection['end'] = group[i]['start'] + step_size

            else:
                # finish the previous merged detection
                if isinstance(group[-1]['score'], list):
                    merged_detection['score'] = [score / consecutives for score in merged_detection['score']]
                else:
                    merged_detection['score'] /= consecutives
                merged_detections.append(merged_detection)

                merged_detection = {'filename': group[i]['filename'], 'start': group[i]['start'], 'end': group[i]['start'] + step_size, 'label': group[i]['label'], 'score': group[i]['score']}
                consecutives = 1

        # finish the last merged detection
        if isinstance(group[-1]['score'], list):
            merged_detection['score'] = [score / consecutives for score in merged_detection['score']]
        else:
            merged_detection['score'] /= consecutives
        merged_detections.append(merged_detection)
    
    return pd.DataFrame(merged_detections)